# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-08-08 10:11:50
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-11 09:28:01

import os
import time
from enum import Enum
from tqdm import tqdm
import numpy as np
import PyCapture2

from .logger import logger
from .constants import *


# Constants
SOFTWARE_TRIGGER = 0x62C
FIRE_VAL = 0x80000000
TRIGGER_INQ = 0x530
CAMERA_POWER = 0x610
POWER_VAL = 0x80000000
CRITICAL_BITRATE = 100  # Critical bitrate (in kb/s) for H264 writing above which some frames are lost
DEFAULT_BITRATE = 80  # Default bitrate (in kb/s) for H264 writing
DEFAULT_JPEG_COMPRESSION = 75  # Default JPEG compression quality (0-100) for MPEG encoding
SHUTTER_TO_FRAME_INTERVAL_RATIO = 0.9  # Ratio of shutter to frame interval to ensure acquisition at specified FPS
CHECK_INTERVAL = 1.  # Interval at which the camera checks for acquisition of interruption (in s)


class TriggerMode(Enum):
    ''' Descriptors of camera trigger modes '''
    STANDARD = 0  # Acquisition starts on trigger and lasts 1 shutter exposure time
    BULB = 1  # Acquisition starts on trigger and lasts trigger width
    SKIP_FRAMES = 3  # Same as STANDARD, but camera transmits 1 out of N frames
    MULTI_EXP_PRESET = 4  # Image readout starts after N expositions lasting each 1 shutter exposure time
    MULTI_EXP_WITDH = 5  # Image readout starts after N expositions lasting each 1 trigger width
    LOW_SMEAR = 13  # Same as STANDARD, but exposure is preceded by CCD fast dump in order to reduce the time each pixel data collects smear
    OVERLAPPED_EXP = 14  # Same as STANDARD, but allows triggering at faster frame rates
    MULTI_SHOT = 15  # Acquire a stream of N images (exposure determined by shutter) upon trigger


class TriggerSource(Enum):
    ''' Descriptors of camera trigger sources '''
    EXTERNAL = 2  # GPIO 2 pin (to adapt to connectics)
    SOFTWARE = 7  # Software


class VideoFormat(Enum):
    ''' Descriptors of video formats '''
    AVI = 0
    MJPG = 1
    H264 = 2


class GrabMode(Enum):
    ''' Descriptors of grab modes '''
    BUFFER_FRAMES = 1
    DROP_FRAMES = 0
    UNSPECIFIED_GRAB_MODE = 2


class CameraError(Exception):
    pass


def log_build_info():
    ''' Log PyCapture build info '''
    lib_ver = PyCapture2.getLibraryVersion()
    logger.info(
        f'PyCapture2 library version: {lib_ver[0]}.{lib_ver[1]}.{lib_ver[2]}.{lib_ver[3]}')


def grab_camera():
    ''' Grab the first detected camera '''
    # Get camera BUS Manager
    bus = PyCapture2.BusManager()
    # Extract number of connected cameras
    ncams = bus.getNumOfCameras()
    # If no camera connected, raise error 
    if ncams == 0:
        raise CameraError(
            'No camera detected. Make sure your camera is connected via a USB3 port')
    # Select camera on 0th index
    cam = Camera()
    cam.connect(bus.getCameraFromIndex(0))
    # Return cam object
    return cam


class Camera(PyCapture2.Camera):
    ''' Interface to FLIR Camera '''

    def __init__(self, *args, **kwargs):
        '''
        Constructor
        
        :param cam: camera object
        '''
        super().__init__(*args, **kwargs)
        self.is_capturing = False
        self.video_stream = None
        self.video_name = None
    
    @property
    def settings(self):
        ''' Get camera settings dictionary '''
        cam_info = self.getCameraInfo()
        return {
            'Serial number' : cam_info.serialNumber,
            'Camera model' : cam_info.modelName.decode('utf-8'),
            'Camera vendor' : cam_info.vendorName.decode('utf-8'),
            'Sensor' : cam_info.sensorInfo.decode('utf-8'),
            'Resolution' : cam_info.sensorResolution.decode('utf-8'),
            'Firmware version' : cam_info.firmwareVersion.decode('utf-8'),
            'Firmware build time' : cam_info.firmwareBuildTime.decode('utf-8'),
            'Frame dimensions': f'{self.ncols} x {self.nrows}',
            'Frame rate': f'{self.get_framerate()} FPS'
        }
    
    def __str__(self):
        ''' String representation: camera model '''
        return self.settings['Camera model']
    
    def connect(self, guid):
        ''' 
        Connect to camera object

        :param guid: Global Unique IDentifier for the camera physical object 
        '''
        logger.info(f'connecting to camera with GUID {guid} ...')
        super().connect(guid)
        # Get frame dimensions 
        self.get_frame_dimensions()
        # Set shutter time < 1 / FPS to enable stable capture at FPS
        self.set_shutter(SHUTTER_TO_FRAME_INTERVAL_RATIO / self.get_framerate() * S_TO_MS)
        # Log settings
        self.log_settings()

    def disconnect(self):
        logger.info(f'disconnecting from {self} camera')
        super().disconnect()

    def log_settings(self):
        ''' Log camera settings '''
        cam_info_str = '\n'.join([f'   - {k}: {v}' for k, v in self.settings.items()])
        logger.info(f'*** CAMERA SETTINGS ***\n{cam_info_str}')
    
    @property
    def config(self):
        ''' Camera configuration '''
        config_prop = self.getConfiguration()
        keys = [
            'asyncBusSpeed ',
            'bandwidthAllocation ', 
            'grabMode ',
            'grabTimeout ', 
            'isochBusSpeed ', 
            'minNumImageNotifications ', 
            'numBuffers ', 
            'numImageNotifications ', 
            'registerTimeout', 
            'registerTimeoutRetries ']
        config_dict = {k.strip(): getattr(config_prop, k) for k in keys}
        config_dict['grabMode'] = GrabMode(config_dict['grabMode'])
        return config_dict

    def log_config(self):
        ''' Log camera configuration '''
        cam_config_str = '\n'.join([f'   - {k}: {v}' for k, v in self.config.items()])
        logger.info(f'*** CAMERA CONFIGURATION ***\n{cam_config_str}')

    def get_grabmode(self):
        '''
        Get frames grab mode
        
        :return: GrabMode enum
        '''
        return self.config['grabMode']
    
    def set_grabmode(self, grabmode):
        ''' 
        Set frames grab mode
        
        :param grabmode: GrabMode enum
        '''
        self.setConfiguration(grabMode=grabmode.value)

    def get_grabtimeout(self):
        ''' 
        Get timeout that will wait for an image before timing out and returning.

        :return: timeout (ms)
        '''
        return self.config['grabTimeout']
    
    def set_grabtimeout(self, timeout=None):
        ''' 
        Set timeout that camera will wait for an image before timing out and returning.

        :param timeout: timeout (ms)
        '''
        if timeout is None: # no timeout case
            timeout = -1
        self.setConfiguration(grabTimeout=timeout)

    def get_framerate(self):
        ''' Get the camera frame rate (in fps) '''
        return self.getProperty(PyCapture2.PROPERTY_TYPE.FRAME_RATE).absValue
    
    def set_framerate(self, fps):
        ''' Set the camera frame rate (in fps) '''
        raise CameraError('Acquisition frame rate cannot be set at the moment')
        # self.setProperty(type=PyCapture2.PROPERTY_TYPE.FRAME_RATE, absValue=fps)

    def get_shutter(self):
        ''' Get camera shutter exposure time (in ms) '''
        return self.getProperty(PyCapture2.PROPERTY_TYPE.SHUTTER).absValue
    
    def set_shutter(self, shutter):
        ''' Set camera shutter exposure time (in ms) '''
        # Ensure that shutter time is compatible with camera frame rate
        fps = self.get_framerate()  # Hz
        if shutter > S_TO_MS / fps:
            raise CameraError(
                f'target shutter time ({shutter:.2f} ms) exceeds inter-frame interval ({S_TO_MS / fps:.2f} ms)')
        logger.info(f'setting shutter time to {shutter:.2f} ms ...')
        self.setProperty(type=PyCapture2.PROPERTY_TYPE.SHUTTER, absValue=shutter)
        s = self.get_shutter()
        logger.info(f'shutter time set to {s:.2f} ms')

    def get_exposure(self):
        ''' Get the automatically adjusted exposure value of the camera '''
        return self.getProperty(PyCapture2.PROPERTY_TYPE.AUTO_EXPOSURE).absValue
    
    def get_temperature(self):
        ''' Get the camera temperature (in dwegrees Celsius) '''
        return self.getProperty(PyCapture2.PROPERTY_TYPE.TEMPERATURE).absValue * 100 - CELSIUS_TO_KELVIN
    
    def get_frame_dimensions(self):
        ''' Get frame dimensions (in pixels) '''
        logger.info('getting frames dimensions...')
        self.disable_trigger(verbose=False)
        self.startCapture(verbose=False)
        img = self.grab_frame()
        self.ncols, self.nrows = img.getCols(), img.getRows()
        self.stopCapture(verbose=False)
    
    def get_format7_settings(self):
        ''' Get format7 settings '''
        fmt7_info, supported = self.getFormat7Info(0)
        return {
            'Supported': supported,
            'Max image pixels': (fmt7_info.maxWidth, fmt7_info.maxHeight),
            'Image unit size': (fmt7_info.imageHStepSize, fmt7_info.imageVStepSize),
            'Offset unit size': (fmt7_info.offsetHStepSize, fmt7_info.offsetVStepSize),
            'Pixel format bitfield': fmt7_info.pixelFormatBitField,
        }
    
    def log_format7_settings(self):
        ''' Log format7 settings '''
        fmt7_settings_dict = self.get_format7_settings()
        fmt7_str = '\n'.join([f'   - {k}: {v}' for k, v in fmt7_settings_dict.items()])
        logger.info(f'*** FORMAT7 SETTINGS ***\n{fmt7_str}')

    def get_image_settings(self):
        ''' Get image settings information dictionary '''
        img_settings, packet_size, pct = self.getFormat7Configuration()
        keys = ['mode', 'width', 'height', 'offsetX', 'offsetY', 'pixelFormat']
        img_settings_dict = {k: getattr(img_settings, k) for k in keys}
        img_settings_dict['packetSize'] = packet_size
        img_settings_dict['packetSizePct'] = pct
        return img_settings_dict

    def log_image_settings(self):
        ''' Log image settings '''
        image_settings_dict = self.get_image_settings()
        image_str = '\n'.join([f'   - {k}: {v}' for k, v in image_settings_dict.items()])
        logger.info(f'*** IMAGE SETTINGS ***\n{image_str}')

    def power_up(self, nretries=10, sleeptime=.1):
        ''' 
        Power up camera

        :param nretries: number of power-up
        :param sleeptime: sleeping time between each attempt (s)
        '''
        # Waiting for Camera to power up
        self.writeRegister(CAMERA_POWER, POWER_VAL)
        for i in range(nretries):
            time.sleep(sleeptime)
            # Camera might not respond to register reads during powerup.
            try:
                reg_val = self.readRegister(CAMERA_POWER)
            except PyCapture2.Fc2error:
                pass
            awake = True
            if reg_val == POWER_VAL:
                break
            awake = False
        if not awake:
            raise CameraError(f'could not wake {self}.')

    def startCapture(self, verbose=True, **kwargs):
        ''' Start video capture and wait for frames to arrive '''
        if verbose:
            logger.info('starting capture')
        super().startCapture(**kwargs)
        self.is_capturing = True
    
    def stopCapture(self, verbose=True):
        ''' Stop video capture '''
        if verbose:
            logger.info('stopping capture')
        super().stopCapture()
        self.is_capturing = False
        # Close video stream and remove incomplete files if they exist
        if self.video_stream is not None:
            logger.info('closing video stream')
            self.video_stream.close()
            if os.path.isfile(self.video_name):
                logger.info(f'deleting incomplete video file {self.video_name}')
                os.remove(self.video_name)
            self.video_stream = None
            self.video_name = None
    
    def grab_frame(self, verbose=False):
        ''' 
        Wrapper around retrieveBuffer that enables checking for camera status
        at regular time intervals (set by grab timeout)
        
        :param verbose: whether to log grab attempts or not
        :return: image object
        '''
        # Execute only if camera acquisition is enabled
        if self.is_capturing:
            # Try to retrieve frame from buffer
            try:
                return self.retrieveBuffer()
            # If PyCapture2 raises error
            except PyCapture2.Fc2error as err:
                # If timeout error, keep trying within recursive call
                if str(err)[2:-2] == 'Timeout error':
                    if verbose:
                        logger.info('camera timed out, re-trying to capture frame ...')
                    return self.grab_frame(verbose=verbose)
                # Otherwise, disable capturing and raise error
                else:
                    self.stopCapture()
                    raise err
       
    def check_software_trigger_presence(self):
        ''' Check if asnychronous trigger is implemented on camera '''
        if self.readRegister(TRIGGER_INQ) & 0x10000 != 0x10000:
            raise (f'SOFT_ASYNC_TRIGGER is not implemented on {self}')
    
    def get_trigger_settings(self):
        ''' Get camera trigger settings dictionary '''
        trigger_settings = self.getTriggerMode()
        trigger_keys = ['mode', 'onOff', 'parameter', 'polarity', 'source']
        trigger_settings_dict = {k: getattr(trigger_settings, k) for k in trigger_keys}
        trigger_settings_dict['mode'] = str(TriggerMode(trigger_settings_dict['mode']))
        trigger_settings_dict['source'] = str(TriggerSource(trigger_settings_dict['source']))
        return trigger_settings_dict

    def log_trigger_settings(self):
        ''' Log camera trigger settings '''
        trigger_settings_dict = self.get_trigger_settings()
        trigger_str = '\n'.join([f'   - {k}: {v}' for k, v in trigger_settings_dict.items()])
        logger.info(f'*** TRIGGER SETTINGS ***\n{trigger_str}')
  
    def set_trigger_settings(self, trigger=True, nframes=0, mode=TriggerMode.MULTI_SHOT,
                             source=TriggerSource.SOFTWARE, verbose=True):
        ''' 
        Configure camera trigger settings
        
        :param trigger: whether to configure the camera to be triggered or not 
        :param nframes: number of frames to capture on trigger (default = infinite)
        :param mode: integer indicating the trigger mode (default = multi-shot)
        :param source: trigger source mode (default = software trigger)
        '''
        # Get trigger settings property
        trigger_settings = self.getTriggerMode()
        # Set trigger ON/OFF
        trigger_settings.onOff = trigger
        # If trigger set to ON, set other properties
        if trigger:
            trigger_settings.mode = mode.value
            trigger_settings.parameter = nframes
            # If trigger source set to software, check that it is enabled on camera
            if source == TriggerSource.SOFTWARE:
                self.check_software_trigger_presence()
            trigger_settings.source = source.value
        # Set trigger settings property
        self.setTriggerMode(trigger_settings)
        if verbose:
            self.log_trigger_settings()
        # If trigger source is software, poll until ready for software trigger
        if trigger and source == TriggerSource.SOFTWARE:
            self.poll_for_software_trigger()

    def wait_for_trigger(self, **kwargs):
        ''' Wrapper around set_trigger_settings with trigger enabled '''
        self.set_trigger_settings(trigger=True, **kwargs)
    
    def disable_trigger(self, **kwargs):
        ''' Wrapper around set_trigger_settings with trigger disabled '''
        self.set_trigger_settings(trigger=False, **kwargs)

    def poll_for_software_trigger(self):
        ''' Poll until the camera is ready for a software trigger '''
        while True:
            reg_val = self.readRegister(SOFTWARE_TRIGGER)
            if not reg_val:
                break
        logger.info(f'{self} is ready for software trigger.')

    def fire_software_trigger(self):
        ''' Fire a software trigger '''
        self.writeRegister(SOFTWARE_TRIGGER, FIRE_VAL)
    
    def open_video_file_stream(self, filename, fmt=VideoFormat.H264, 
                               bitrate=DEFAULT_BITRATE, jpeg_compression=75):
        '''
        Open video stream to file
        
        :param fname: name of file to stream to
        :param fmt: video format (default = H264)
        :param bitrate: H264 bitrate in kb/s (default = 200)
        :param jpeg_compression: JPEG compression quality (1-100, default = 75)
        :return: video stream object
        '''
        # Convert filename to UTF8
        fname = filename.encode('utf-8')
        # Get camera frame rate
        fps = self.get_framerate()
        # Create video object and appropriate file stream
        video = PyCapture2.FlyCapture2Video()
        if fmt == VideoFormat.AVI:
            video.AVIOpen(fname, fps)
        elif fmt == VideoFormat.MJPG:
            video.MJPGOpen(fname, fps, jpeg_compression)
        elif fmt == VideoFormat.H264:
            if bitrate > CRITICAL_BITRATE:
                raise CameraError(
                    f'specified H.264 encoding bitrate ({bitrate} kb/s) will likely induce some frame loss')
            video.H264Open(fname, fps, self.ncols, self.nrows, bitrate * KBS_TO_BS)
        else:
            raise CameraError(f'{fmt} video format not available.')
        return video

    def save_video_to_file(self, filename, nframes, verbose=False, **kwargs):
        '''
        Save current video acquisition to file
        
        :param filename: output filename
        :param nframes: number of frames to save 
        '''
        # Open video stream to output file
        self.video_stream = self.open_video_file_stream(filename, **kwargs)
        self.video_name = filename
        # Log start
        logger.info(f'saving {nframes} frames to {filename} ...')
        # Loop until all frames have been acquired
        ngrabbed = 0
        pbar = None
        while ngrabbed < nframes:
            try:
                # Grab frame
                img = self.grab_frame(verbose=verbose)
                # Append image to video stream (only if True image object was returned)
                if img is not None:
                    self.video_stream.append(img)
                # On first frame, create progress bar and set acquisition start
                if pbar is None:
                    tstart = time.perf_counter()
                    pbar = tqdm(total=nframes)
                # Increment progress bar and ngrabbed counter
                ngrabbed += 1
                pbar.update()
            except PyCapture2.Fc2error as fc2Err:
                logger.error(f'error retrieving buffer: {str(fc2Err)}')
                continue
        # Close progress bar and stop acquisition time
        tacq = time.perf_counter() - tstart
        pbar.close()
        # Log acquisition time
        logger.info(f'acquisition time: {tacq:.2f} s')
        # Close video object
        self.video_stream.close()
        self.video_stream = None
        self.video_name = None

    def acquire(self, filename, duration, nacqs=1, 
                trigger_source=TriggerSource.SOFTWARE, verbose=False, **kwargs):
        '''
        Acquire video(s) and save to file(s)
        
        :param filename: output filename(s)
        :param duration: duration of each acquisition (in s)
        :param nacqs: number of successive acquisitions
        :param trigger source: trigger source (default = software)
        '''
        # Determine output file names
        if nacqs > 1:
            root, ext = os.path.splitext(filename)
            fnames = [f'{root}_{i + 1:05}{ext}' for i in range(nacqs)]
        else:
            fnames = [filename]
        # Determine number of frames to save according to specified duration & FPS
        nframes = int(np.ceil(duration * self.get_framerate()))
        # Ensure full capture even in case of frame loss
        nframes_capture = nframes # int(np.ceil(nframes * 1.1))  
        logger.info(f'number of frames per acquisition: {nframes}')
        # Set frames grabbing settings
        self.set_grabtimeout(CHECK_INTERVAL * S_TO_MS)
        self.set_grabmode(GrabMode.BUFFER_FRAMES)
        # Wait for appropriate trigger
        self.wait_for_trigger(nframes=nframes_capture, source=trigger_source)
        # For each acquisition
        for iacq, fname in enumerate(fnames):
            # Start capture (& open buffer)
            self.startCapture(verbose=verbose)
            # If trigger source set to software, fire trigger
            if trigger_source == TriggerSource.SOFTWARE:
                self.fire_software_trigger()
            # Acquire video and save to file
            logger.info(f'starting acqusition {iacq + 1}/{nacqs}')
            self.save_video_to_file(fname, nframes, verbose=verbose, **kwargs)
            # Stop capture (& clear buffer)
            self.stopCapture(verbose=verbose)
    


# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-03-15 15:44:20
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-17 10:54:14

''' Initiate test sequence with Keysight network analyzer. '''

import logging

from instrulink import VisaError, logger
from instrulink.keysight_e5061b import KeysightE5061B


logger.setLevel(logging.DEBUG)

try:
    # Grab function generator
    instrument = KeysightE5061B()
    print(instrument)

    instrument.get_channel_data(ich=1)

except VisaError as e:
    logger.error(e)
    quit()


# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-04-07 17:51:29
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-17 10:55:01

import argparse
import matplotlib.pyplot as plt

from instrulink import logger, si_format, grab_oscilloscope, VisaError
from instrulink.constants import TTL_PAMP

# Default acquisition parameters
tscale = 1e-3  # temporal scale (s/div)
vscale = 0.2  # vertical scale (V/div)
voffset = 0.  # vertical offset (V)
tdelay = 5e-3 # 3e-3  # trigger delay (s)

# Parse command line arguments
parser = argparse.ArgumentParser()
parser.add_argument(
    '--model', type=str, default=None, help='Oscilloscope model')
parser.add_argument(
    '--isignal', type=int, default=1, choices=(1, 2, 3, 4), help='Signal channel index')
parser.add_argument(
    '--itrigger', type=int, default=1, choices=(1, 2, 3, 4), help='Trigger channel index')
args = parser.parse_args()
ich_sig = args.isignal  # signal channel index
ich_trig = args.itrigger  # trigger channel index
ichs = set([ich_sig, ich_trig]) # set of channel indices

try:
    # Grab oscilloscope object
    scope = grab_oscilloscope(key=args.model)

    # Display settings
    print('DISPLAY SETTINGS')
    scope.display_menu()
    for ich in ichs:
        scope.show_trace(ich)
        logger.info(f'trace on channel {ich}: {scope.is_trace(ich)}')

    # Probes & coupling settings
    print('PROBE & COUPLING SETTINGS')
    for ich in ichs:
        scope.set_probe_attenuation(ich, 1)
        logger.info(f'channel {ich} attenuation factor = {scope.get_probe_attenuation(ich)}')
        logger.info(f'channel {ich} coupling mode: {scope.get_coupling_mode(ich)}')
    
    # Temporal and vertical scales / offsets
    print('SCALE & OFFSET SETTINGS')
    # scope.auto_setup()
    scope.set_temporal_scale(tscale)
    logger.info(f'temporal scale = {scope.get_temporal_scale()} s/div')
    for ich in ichs:
        scope.set_vertical_offset(ich, voffset)
        logger.info(f'channel {ich} vertical offset = {scope.get_vertical_offset(ich)} V')
    scope.set_vertical_scale(ich_sig, vscale)
    logger.info(f'channel {ich_sig} vertical scale = {scope.get_vertical_scale(ich_sig)} V/div')
    if ich_sig != ich_trig:
        scope.set_vertical_scale(ich_trig, TTL_PAMP / 2)
        logger.info(f'channel {ich_trig} vertical scale = {scope.get_vertical_scale(ich_trig)} V/div')

    # Trigger settings
    print('TRIGGER SETTINGS')
    scope.set_trigger_mode('norm')
    logger.info(f'trigger mode = {scope.get_trigger_mode()}')
    scope.set_trigger_coupling_mode(ich_trig, 'DC')
    logger.info(f'channel {ich_trig} trigger coupling mode = {scope.get_trigger_coupling_mode(ich_trig)}')
    logger.info(f'trigger type = {scope.get_trigger_type()}')
    scope.set_trigger_source(ich_trig)
    logger.info(f'trigger source channel = {scope.get_trigger_source()}')
    scope.set_trigger_slope(ich_trig, 'POS')
    logger.info(f'channel {ich_trig} trigger slope = {scope.get_trigger_slope(ich_trig)}')
    if ich_sig == ich_trig:
        scope.set_trigger_level(ich_trig, vscale / 2)
    else:
        scope.set_trigger_level(ich_trig, TTL_PAMP / 2)
    logger.info(f'channel {ich_trig} trigger level = {scope.get_trigger_level(ich_trig)} V')
    scope.set_trigger_delay(scope.get_temporal_range() / 3)
    logger.info(f'trigger delay = {si_format(scope.get_trigger_delay())}s')

    # Cursor settings
    print('CURSORS SETTINGS')
    if args.model == 'bk':
        for ich in ichs:
            for ctype in scope.CURSOR_TYPES:
                cpos = scope.get_cursor_position(ich, ctype)
                logger.info(f'channel {ich} {ctype} cursor position: {cpos} divs')
            for ctype in scope.CVALUES_TYPES:
                cval = scope.get_cursor_value(ich, ctype)
                logger.info(f'channel {ich} {ctype} values: {cval}')

    # Filter settings
    print('FILTER SETTINGS')
    scope.enable_bandwith_filter(ich_sig)
    scope.disable_bandwith_filter(ich_sig)
    sfilt = 'enabled' if scope.is_bandwith_filter_enabled(ich_sig) else 'disabled'
    logger.info(f'bandwidth filter for channel {ich_sig} is {sfilt}')
    fc = 10 / scope.get_temporal_scale()  # cutoff frequency (Hz)
    scope.set_filter(ich_sig, 'LP', fhigh=fc)
    scope.enable_filter(ich_sig)
    sfilt = 'enabled' if scope.is_filter_enabled(ich_sig) else 'disabled'
    logger.info(f'channel {ich_sig} filter is {sfilt}')

    # Acquisition settings
    print('ACQUISITION SETTINGS')
    logger.info(f'sample rate = {si_format(scope.get_sample_rate())}Hz')
    logger.info(f'acquisition type: {scope.get_acquisition_type()}')
    if scope.get_acquisition_type().startswith('AVER'):
        logger.info(f'# sweeps / acq: {scope.get_nsweeps_per_acquisition()}')
    scope.set_trigger_mode('norm')
    logger.info(f'trigger mode = {scope.get_trigger_mode()}')
    logger.info(f'trigger type = {scope.get_trigger_type()}')

    # Waveform settings
    print('WAVEFORM PARAMETERS')
    for k, v in scope.UNITS_PER_PARAM.items():
        try:
            val = scope.get_parameter_value(ich_sig, k)
            logger.info(f'{k} = {val} {v}')
        except VisaError as e:
            logger.error(e)
    
    # Hide menu before acquisition
    scope.hide_menu()

    # Waveform data
    print('WAVEFORM DATA')
    fig = scope.plot_waveform_data(ich_sig, n=5)

    # Screen data
    print('SCREEN CAPTURE')
    fig = scope.plot_screen_capture()

    # Show graphs
    plt.show()

except VisaError as e:
    logger.error(e)



# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-04-27 18:30:00
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-17 10:54:43

from instrulink import grab_manipulator, SutterError, logger

try:
    # Grab micro-manipulator
    mp = grab_manipulator()

    # Display position, velocity and resolution
    mp.get_position(verbose=True)
    mp.get_velocity(verbose=True)
    mp.get_resolution(verbose=True)

    # Change velocity
    mp.set_velocity(1000)
    mp.get_velocity(verbose=True)

    # Change motion mode
    mp.set_resolution(0)
    mp.get_resolution(verbose=True)
    
    # Move up 1 mm
    mp.translate([0, 0, -1000])
    mp.get_position(verbose=True)

    mp.disconnect()

except SutterError as e:
    logger.error(e)
    quit()


# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-04-07 17:51:29
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-08-08 12:39:13
# @Last Modified time: 2022-04-08 21:17:22

import abc
import re
import io
from PIL import Image
import matplotlib.pyplot as plt

from .constants import *
from .si_utils import *
from .visa_instrument import *


class Oscilloscope(VisaInstrument):
    ''' Base class for oscilloscopes. '''
    
    # --------------------- REQUIRED ATTRIBUTES ---------------------
    
    @property
    @abc.abstractmethod
    def CHANNELS(self):
        ''' Available channel indexes '''
        raise NotImplementedError
    
    @property
    @abc.abstractmethod
    def NHDIVS(self):
        ''' Number of horizontal divisions '''
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def NVDIVS(self):
        ''' Number of vertical divisions '''
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def MAX_VDIV(self):
        ''' Max voltage per division (V) '''
        raise NotImplementedError
    
    @property
    @abc.abstractmethod
    def COUPLING_MODES(self):
        ''' Available coupling modes '''
        raise NotImplementedError
    
    @property
    @abc.abstractmethod
    def ACQ_TYPES(self):
        ''' Acquisition types '''
        raise NotImplementedError
        
    @property
    @abc.abstractmethod
    def NAVGS(self):
        ''' Number of samples to average for average acquisition '''
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def TRIGGER_MODES(self):
        ''' Available trigger modes '''
        raise NotImplementedError
    
    @property
    @abc.abstractmethod
    def TRIGGER_COUPLING_MODES(self):
        ''' Available trigger coupling modes '''
        raise NotImplementedError
    
    @property
    @abc.abstractmethod
    def TRIG_TYPES(self):
        ''' Available trigger types '''
        raise NotImplementedError
    
    @property
    @abc.abstractmethod
    def TRIGGER_SLOPES(self):
        ''' Available trigger slopes '''
        raise NotImplementedError
    
    @property
    @abc.abstractmethod
    def INTERP_TYPES(self):
        ''' Available interpolation types '''
        raise NotImplementedError
    
    @property
    @abc.abstractmethod
    def UNITS_PER_PARAM(self):
        ''' Units per parameter dictionary '''
        raise NotImplementedError
    
    # --------------------- MISCELLANEOUS ---------------------

    @abc.abstractmethod
    def calibrate(self):
        ''' Calibrate oscilloscope '''
        raise NotImplementedError
    
    # --------------------- DISPLAY ---------------------

    @abc.abstractmethod
    def display_menu(self):
        ''' Display menu '''
        raise NotImplementedError

    @abc.abstractmethod
    def hide_menu(self):
        ''' Hide menu '''
        raise NotImplementedError

    @abc.abstractmethod
    def show_trace(self, ich):
        ''' Enable trace display on specific channel '''
        raise NotImplementedError

    @abc.abstractmethod
    def hide_trace(self, ich):
        ''' Disable trace display on specific channel '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def is_trace(self, ich):
        ''' Query trace display on specific channel '''
        raise NotImplementedError
    
    def restrict_traces(self, ichs):
        ''' Restrict trace display to specific channels '''
        # Show traces for speficied channels
        for ich in ichs:
            self.show_trace(ich)
        # Hide traces for all other channels
        for ich in list(set(self.CHANNELS) - set(ichs)):
            if self.is_trace(ich):
                self.hide_trace(ich)
    
    @abc.abstractmethod
    def get_screen_binary_img(self):
        ''' Extract screen capture image in binary format '''
        raise NotImplementedError

    def capture_screen(self):
        '''
        Capture screen information in image format
        
        :return: PIL image object representing the screenshot
        '''
        # Extract binary image
        binary_img = self.get_screen_binary_img()
        # Convert image to PIL readable object and return
        return Image.open(io.BytesIO(binary_img))

    # --------------------- SCALES / OFFSETS ---------------------

    @abc.abstractmethod
    def auto_setup(self):
        ''' Perform auto-setup. '''
        raise NotImplementedError

    @abc.abstractmethod
    def set_temporal_scale(self, value):
        ''' Set the temporal scale (in s/div) '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def get_temporal_scale(self):
        ''' Get the temporal scale (in s/div) '''
        raise NotImplementedError
    
    def get_temporal_range(self):
        ''' Get temporal display range at current temporal scale range (in s) '''
        return self.get_temporal_scale() * self.NHDIVS
    
    @abc.abstractmethod
    def set_vertical_scale(self, ich, value):
        ''' Set the vertical sensitivity of the specified channel (in V/div) '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_vertical_scale(self, ich):
        ''' Get the vertical sensitivity of the specified channel (in V/div) '''
        raise NotImplementedError

    @abc.abstractmethod
    def set_vertical_offset(self, ich, value):
        ''' Set the vertical offset of the specified channel (in V) '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_vertical_offset(self, ich):
        ''' Get the vertical offset of the specified channel (in V) '''
        raise NotImplementedError
    
    # --------------------- FILTERS ---------------------

    @abc.abstractmethod
    def enable_bandwith_filter(self, ich):
        ''' Turn on bandwidth-limiting low-pass filter for specific channel '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def disable_bandwith_filter(self, ich):
        ''' Turn off bandwidth-limiting low-pass filter for specific channel '''
        raise NotImplementedError

    @abc.abstractmethod    
    def is_bandwith_filter_enabled(self, ich):
        ''' Query bandwidth-limiting low-pass filter for specific channel '''
        raise NotImplementedError
    
    def set_filter(self, *args, **kwargs):
        logger.warning('no filter capabilities')
    
    def enable_filter(self, ich):
        logger.warning('no filter capabilities')

    def is_filter_enabled(self, ich):
        return False
    
    # --------------------- PROBES & COUPLING ---------------------

    @abc.abstractmethod
    def get_probe_attenuation(self, ich):
        ''' Get the vertical attenuation factor of a specific channel '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def set_probe_attenuation(self, ich, value):
        ''' Set the vertical attenuation factor of a specific channel '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def get_coupling_mode(self, ich):
        ''' Get the coupling mode of a specific channel '''
        raise NotImplementedError

    @abc.abstractmethod
    def set_coupling_mode(self, ich, value):
        ''' Set the coupling mode of a specific channel '''
        raise NotImplementedError

    # --------------------- TRIGGER ---------------------

    @abc.abstractmethod
    def get_trigger_mode(self):
        ''' Get trigger mode '''
        raise NotImplementedError
     
    @abc.abstractmethod 
    def set_trigger_mode(self, value):
        ''' Set trigger mode '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_trigger_coupling_mode(self, *args, **kwargs):
        ''' Get trigger coupling mode. '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def set_trigger_coupling_mode(self, *args, **kwargs):
        ''' Set trigger coupling mode. '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def get_trigger_type(self):
        ''' Get trigger type '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def set_trigger_type(self, val):
        ''' Set trigger type '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def get_trigger_source(self):
        ''' Get trigger source channel index '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def set_trigger_source(self, ich):
        ''' Set trigger source channel index '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_trigger_level(self, *args, **kwargs):
        ''' Get trigger level (in V) '''
        raise NotImplementedError

    @abc.abstractmethod
    def set_trigger_level(self, *args, **kwargs):
        ''' Set trigger level (in V) '''
        raise NotImplementedError

    def check_trigger_delay(self, value):
        ''' Check that trigger delay value falls within current temporal range '''
        thalfrange = self.get_temporal_range() / 2
        if np.abs(value) > thalfrange:
            logger.warning(
                f'target temporal delay ({si_format(value, 2)}s) outside of current display temporal bounds (+/- {si_format(thalfrange, 2)}s) -> restricting')
            value = np.sign(value) * thalfrange
        return value
    
    @abc.abstractmethod
    def get_trigger_delay(self):
        ''' Get trigger delay (in s) '''
        raise NotImplementedError

    @abc.abstractmethod
    def set_trigger_delay(self, value):
        ''' Set trigger delay (in s) '''
        raise NotImplementedError
    
    def set_relative_trigger_delay(self, value):
        ''' Set trigger delay as a fraction of the current temporal scale '''
        tdiv = self.get_temporal_scale()
        self.set_trigger_delay(value * tdiv) 

    @abc.abstractmethod
    def force_trigger(self):
        ''' Force the instrument to make 1 acquisition '''
        raise NotImplementedError
    
    # --------------------- ACQUISITION ---------------------
    
    @abc.abstractmethod
    def arm_acquisition(self):
        '''
        Enables the signal acquisition process by changing the acquisition state
        (trigger mode) from "stopped" to "single".
        '''
        raise NotImplementedError
    
    def stop_acquisition(self):
        '''
        Immediately stops the acquisition of a signal (if the trigger mode is
        AUTO or NORM).
        '''
        self.write('STOP')

    @abc.abstractmethod
    def get_nsweeps_per_acquisition(self):
        ''' Get the number of samples to average from for average acquisition.'''
        raise NotImplementedError

    @abc.abstractmethod    
    def set_nsweeps_per_acquisition(self, value):
        ''' Set the number of samples to average from for average acquisition.'''
        raise NotImplementedError
    
    def get_acquisition_status(self):
        ''' Get the acquisition status of the oscilloscope '''
        out = self.query('SAST?')
        mo = re.match('SAST (.+)', out)
        return mo[1]
    
    @abc.abstractmethod
    def get_sample_rate(self):
        ''' Get the acquisition sampling rate (in samples/second) '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def get_nsamples(self, *args, **kwargs):
        ''' Get the number of samples per acquisition '''
        raise NotImplementedError

    @abc.abstractmethod
    def enable_peak_detector(self):
        ''' Turn on peak detector '''
        raise NotImplementedError

    @abc.abstractmethod
    def disable_peak_detector(self):
        ''' Turn off peak detector '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_acquisition_type(self):
        ''' Get oscilloscope acquisition type '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def set_acquisition_type(self, value):
        ''' Set oscilloscope acquisition type '''
        raise NotImplementedError
    
    # --------------------- PARAMETERS ---------------------

    @abc.abstractmethod
    def get_parameter_value(self, ich, pkey):
        ''' Query the value of a parameter on a particular channel '''
        raise NotImplementedError
    
    # --------------------- WAVEFORMS ---------------------

    def set_waveform_settings(self, *args, **kwargs):
        ''' Set waveform settings '''
        pass

    @abc.abstractmethod
    def get_waveform_data(self, ich):
        '''
        Get waveform data from a specific channel
        
        :param ich: channel index
        :return: 2-tuple of numpy arrays with:
            - t: time signal (s)
            - y: waveform signal (V)
        '''
        raise NotImplementedError
    
    # --------------------- PLOTTING ---------------------
    
    def plot_waveform_data(self, *args, n=1, **kwargs):
        '''
        Acquire and plot waveform data
        
        :param n: number of acquisitions to plot
        :return: figure handle
        '''
        fig, ax = plt.subplots()
        for k in ['right', 'top']:
            ax.spines[k].set_visible(False)
        ax.set_xlabel('time (ms)')
        ax.set_ylabel('voltage (V)')
        ax.set_title('waveform data')
        for i in range(n):
            t, y = self.get_waveform_data(*args, **kwargs)
            ax.plot(t * 1e3, y, label=f'acq{i + 1}')
        ax.axvline(0, c='k', ls='--')
        ax.axhline(0, c='k', ls='--')
        if n > 1:
            ax.legend()
        return fig
    
    def plot_screen_capture(self):
        '''
        Acquire and render screen capture
        
        :return: figure handle
        '''
        img = self.capture_screen()
        fig, ax = plt.subplots()
        ax.imshow(img)
        ax.axis('off')
        return fig
    
    # --------------------- MULTI-CHANNEL SETTING ---------------------

    def set_multichannel_vscale(self, vscale, ich_signal, ich_trigger=None, trig_detect=None):
        ''' 
        Set scope vertical scale and trigger settings 
        
        :param vdiv: vertical scale of signal channel (V/div)
        :param ich_signal: signal channel index
        :param ich_trigger: trigger channel index (optional)
        :param trig_detect: trigger detection amplitude (defaults to TTL_PAMP / 2)
        '''
        # Set trigger detection amplitude if not specified
        if trig_detect is None:
           trig_detect = TTL_PAMP / 2
        
        # Set vertical scale on signal channel
        self.set_vertical_scale(ich_signal, vscale)

        # Set vertical scale & trigger level on trigger channel, if any
        if ich_trigger is not None:
            self.set_vertical_scale(ich_trigger, trig_detect)
            self.set_trigger_level(ich_trigger, trig_detect)

        # Set trigger source to appropriate channel
        self.set_trigger_source(ich_trigger if ich_trigger is not None else ich_signal)
        



# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-03-15 09:26:06
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-11 16:57:22

import abc
import pyvisa
import time
import re
import threading
import numpy as np

from .constants import S_TO_MS
from .logger import logger
from .si_utils import si_format, SI_powers


class VisaError(Exception):
    ''' Custom exception class for VISA instrument '''
    pass


def list_visa_resources():
    ''' List all available VISA resources. '''
    rm = pyvisa.ResourceManager()
    resources = rm.list_resources()
    res_str = '\n'.join([f'  - {r}' for r in resources])
    print(f'VISA resources:\n{res_str}')


class VisaInstrument(metaclass=abc.ABCMeta):
    ''' Generic interface to a VISA instrument using the SCPI command interface '''

    PREFIX = ''  # Prefix to add to each command
    _lock = threading.Lock()  # Lock to prevent concurrent access to the instrument

    def __init__(self, testmode=False, lock=False):
        ''' Initialization. '''
        self.instrument_handle = None
        self.testmode = testmode
        self.lock = lock
        if not testmode:
            self.connect()

    def __repr__(self):
        ''' String representation '''
        return f'{self.__class__.__name__}: {self.get_idn()}'

    def __del__(self):
        ''' Destruction. '''
        if self.is_connected():
            logger.info(f'releasing {self.__class__.__name__} resource')
    
    # --------------------- MISCELLANEOUS ---------------------

    def connect(self):
        ''' Connect to instrument. '''
        # Detect instrument and raise error if not found
        rm = pyvisa.ResourceManager()
        resources = rm.list_resources()
        if len(resources) == 0:
            raise VisaError('no instrument detected')
        res_id = next((item for item in resources if re.search(self.USB_ID, item) is not None), None)
        if res_id is None:
            raise VisaError(
                f'{self.__class__.__name__} instrument ID "{self.USB_ID}" not detected in USB resources.'
                ' Check the USB connection or update the instrument USB ID.')

        # Open resource and store its handle
        self.instrument_handle = rm.open_resource(res_id)

        # Reset instrument and clear error queue
        self.reset()
        self.clear()

        # Set instrument timeout, if defined
        if hasattr(self, 'TIMEOUT_SECONDS'):
            self.timeout = self.TIMEOUT_SECONDS * S_TO_MS
        
        # Print connection info
        print(f'{f" {repr(self)} ":-^100}')

    def disconnect(self):
        ''' Disconnect from instrument. '''
        self.instrument_handle = None

    def is_connected(self):
        ''' Check if instrument is connected. '''
        return self.instrument_handle is not None
    
    def process_text(self, text):
        ''' Process text before sending to instrument. '''
        if not text.startswith('*'):
            return f'{self.PREFIX}{text}'
        return text
    
    def query(self, text):
        ''' Query instrument and return response. '''
        if self.lock:
            self._lock.acquire()
        text = self.process_text(text)
        logger.debug(f'QUERY: {text}')
        if not self.testmode:
            out = self.instrument_handle.query(text)[:-1]
        else: 
            out = None
        if self.lock:
            self._lock.release()
        return out
    
    def query_binary_values(self, text, *args, **kwargs):
        ''' Query instrument and return binary response. '''
        if self.lock:
            self._lock.acquire()
        text = self.process_text(text)
        logger.debug(f'QUERY_BINARY_VALUES: {text}')
        if not self.testmode:
            out = self.instrument_handle.query_binary_values(text, *args, **kwargs)
        else:
            out = None
        if self.lock:
            self._lock.release()
        return out

    def write(self, text):
        ''' Send command to instrument. '''
        if self.lock:
            self._lock.acquire()
        text = self.process_text(text)
        logger.debug(f'WRITE: {text}')
        if not self.testmode:
            self.instrument_handle.write(f'{text}')
        if self.lock:
            self._lock.release()
    
    def read_raw(self):
        ''' Read the raw binary content of an instrument answer '''
        if self.lock:
            self._lock.acquire()
        out = self.instrument_handle.read_raw()
        if self.lock:
            self._lock.release()
        return out

    def reset(self):
        ''' Reset the function generator to its factory default state '''
        self.write('*RST')
    
    def clear(self):
        ''' Clear the error queue. '''
        self.write('*CLS')
    
    def get_idn(self):
        '''
        Get instrument ID string.

        :return: string consisting of 4 parts separated by commas:
            - the 1st part is the manufacturer name
            - the 2nd part is the instrument model name
            - the 3rd part is the instrument serial number
            - the 4th part is the digital board version number or some other
            information about the instrument
        '''
        try:
            return self.query('*IDN?')
        except pyvisa.errors.VisaIOError as e:
            return None
    
    def get_name(self):
        ''' Get a simplified manufacturer-model string representation of the instrument '''
        idn = self.get_idn()
        if idn is None:
            return None
        else:
            manufacturer, model, _, _ = idn.split(',')
            return f'{manufacturer} {model}'
    
    def is_operation_complete(self):
        ''' Query whether the previous operations are completed. '''
        return bool(int(self.query('*OPC?')))
    
    abc.abstractmethod
    def wait(self, *args, **kwargs):
        ''' Wait for previous command to finish. '''
        raise NotImplementedError

    @property
    def timeout(self):
        ''' Get I/O operations timeout paramerer (in ms) '''
        return self.instrument_handle.timeout
    
    @timeout.setter
    def timeout(self, value):
        ''' Set I/O operations timeout paramerer (in ms) '''
        self.instrument_handle.timeout = value

    # --------------------- I/O PROCESSING ---------------------

    def si_process(self, val):
        ''' Process an input value to be set '''
        return si_format(val, space='').upper()
    
    def process_int_mo(self, out, rgxp, key):
        ''' Process an integer notation regexp match object '''
        mo = re.match(rgxp, out)
        if mo is None:
            raise VisaError(f'could not extract {key} from "{out}" (rgxp = "{rgxp}")')
        return int(mo[1])
    
    def process_float(self, val, suffix):
        if suffix in self.UNITS:
            exp = 0
        else:
            try:
                exp = SI_powers[suffix[0]]
            except KeyError:
                exp = SI_powers[suffix[0].swapcase()]
        factor = np.float_power(10, exp)
        return val * factor
    
    def process_float_mo(self, out, rgxp, key):
        ''' Process a float notation regexp match object '''
        mo = re.match(rgxp, out)
        if mo is None:
            raise VisaError(f'could not extract {key} from "{out}" (rgxp = "{rgxp}")')
        val = float(mo[1])
        suffix = mo[2]
        return self.process_float(val, suffix)
    
    # --------------------- ERRORS ---------------------

    @property
    @abc.abstractmethod
    def NO_ERROR_CODE(self):
        ''' String returned when no error has occurred '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_last_error(self):
        ''' Get last entry in error queue. '''
        raise NotImplementedError
    
    def check_error(self):
        ''' Check if error was generated. '''
        err_msg = self.get_last_error()
        if err_msg != self.NO_ERROR_CODE:
            raise VisaError(err_msg)
        
    # --------------------- FRONT PANEL & DISPLAY ---------------------

    @abc.abstractmethod
    def lock_front_panel(self):
        ''' Lock the front panel of the instrument.'''
        raise NotImplementedError

    @abc.abstractmethod
    def unlock_front_panel(self):
        ''' Unlock the front panel of the instrument.'''
        raise NotImplementedError
    
    #--------------------- CHAINING ---------------------

    def chain(self, func1, func2, interval=1.):
        ''' Call 2 functions separated by a given interval (in s). '''
        if interval <= 0:
            raise VisaError(f'interval must be strictly positive')
        def func2wrap():
            time.sleep(interval)
            func2()
        func1()
        if interval == 0.:
            func2()
        else:
            thread = threading.Thread(target=func2wrap)
            thread.start()
    
    #--------------------- TRIGGER ---------------------

    def trigger(self):
        ''' Trigger the instrument. '''
        self.write('*TRG')

    @abc.abstractmethod
    def set_trigger_slope(self, *args, **kwargs):
        ''' Select the trigger slope. '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_trigger_slope(self, *args, **kwargs):
        ''' Get the trigger slope. '''
        raise NotImplementedError

    #--------------------- CHANNELS ---------------------

    def check_channel_index(self, ich):
        ''' Check if channel index is valid. '''
        if ich not in self.CHANNELS:
            raise VisaError(f'{ich} is not a valid channel index (values are {self.CHANNELS})')

    #--------------------- DATA TRANSFER ---------------------

    @property
    def chunk_size(self):
        ''' Get data transfer chunk size (in bytes) '''
        return self.instrument_handle.chunk_size

    @chunk_size.setter
    def chunk_size(self, value):
        ''' Set data transfer chunk size (in bytes) '''
        self.instrument_handle.chunk_size = value


# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Email: theo.lemaire@epfl.ch
# @Date:   2021-02-21 23:28:05
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-22 09:52:47

import logging
from functools import wraps

from instrulink import logger, VisaError
from instrulink.keysight_33500b import Keysight33500B

logger.setLevel(logging.DEBUG)


def check_error(testfunc):
    @wraps(testfunc)
    def wrapper(self, *args, **kwargs):
        try:
            out = testfunc(self, *args, **kwargs)
            self.instrument.check_error()
            return out
        except VisaError as err:
            logger.error(f'ERROR: {err}')
    return wrapper


class Keysight33500BTest:

    # Sine wave parameters
    Fdrive = 500e3    # Hz
    Vpp = 0.06        # Vpp
    phi = 0.          # degrees
    duration = 80e-6  # s

    def __init__(self, inst):
        ''' Initialization '''
        # Create function generator object
        self.instrument = inst

    @check_error
    def continous_wave(self):
        ''' Apply continous sine wave. '''
        choice = input('apply continous wave (y/n)?:')
        if choice == 'y':
            self.instrument.apply_continous_wave('SIN', self.Fdrive, self.Vpp)
            input('press any key to stop')
            self.instrument.disable_output()

    @check_error
    def single_pulse(self):
        ''' Set sine pulse parameters. '''
        choice = input('set sine pulse (y/n)?:')
        if choice == 'y':
            self.instrument.set_sine_pulse(self.Fdrive, self.Vpp, self.phi, self.duration)
            stop = False
            while not stop:
                choice = input('press SPACE send pulse or "s" to stop:')
                if choice == ' ':
                    self.instrument.send_pulse()
                elif choice == 's':
                    stop = True

    def all(self):
        self.continous_wave()
        self.single_pulse()


try:
    inst = Keysight33500B()
    tester = Keysight33500BTest(inst)
    # tester.all()
except VisaError as err:
    logger.error(f'ERROR: {err}')



# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-08-15 09:55:23
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-11 15:34:28

from .factory import *
from .logger import *
from .si_utils import *
from .visa_instrument import VisaError, list_visa_resources
from .sutter_mp285a import SutterError


# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-04-07 19:21:50
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2022-08-15 09:59:04

import numpy as np
import operator

# SI units prefixes
SI_powers = {
    'y': -24,  # yocto
    'z': -21,  # zepto
    'a': -18,  # atto
    'f': -15,  # femto
    'p': -12,  # pico
    'n': -9,   # nano
    'u': -6,   # micro
    'm': -3,   # mili
    '': 0,     # None
    'k': 3,    # kilo
    'M': 6,    # mega
    'G': 9,    # giga
    'T': 12,   # tera
    'P': 15,   # peta
    'E': 18,   # exa
    'Z': 21,   # zetta
    'Y': 24,   # yotta
}
si_prefixes = {k: np.power(10., v) for k, v in SI_powers.items()}
sorted_si_prefixes = sorted(si_prefixes.items(), key=operator.itemgetter(1))


def get_SI_pair(x, scale='lin', unit_dim=1):
    ''' Get the correct SI factor and prefix for a floating point number. '''
    if isinstance(x, (list, tuple, np.ndarray)):
        # If iterable, get a representative number of the distribution
        x = np.asarray(x)
        x = x.prod()**(1.0 / x.size) if scale == 'log' else np.mean(x)
    if x == 0:
        return 1e0, ''
    else:
        vals = np.array([tmp[1] for tmp in sorted_si_prefixes])
        if unit_dim != 1:
            vals = np.power(vals, unit_dim)
        ix = np.searchsorted(vals, np.abs(x)) - 1
        if np.abs(x) == vals[ix + 1]:
            ix += 1
        return vals[ix], sorted_si_prefixes[ix][0]


def si_format(x, precision=0, space=' ', **kwargs):
    ''' Format a float according to the SI unit system, with the appropriate prefix letter. '''
    if isinstance(x, float) or isinstance(x, int) or isinstance(x, np.float) or\
       isinstance(x, np.int32) or isinstance(x, np.int64):
        factor, prefix = get_SI_pair(x, **kwargs)
        return f'{x / factor:.{precision}f}{space}{prefix}'
    elif isinstance(x, list) or isinstance(x, tuple):
        return [si_format(item, precision, space) for item in x]
    elif isinstance(x, np.ndarray) and x.ndim == 1:
        return [si_format(float(item), precision, space) for item in x]
    else:
        raise ValueError(f'cannot si_format {type(x)} objects')



# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-03-15 09:26:06
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-08-07 16:14:09

import abc
import numpy as np

from .visa_instrument import *


class WaveformGenerator(VisaInstrument):
    ''' Generic interface to a waveform generator instrument using the SCPI command interface '''

    PREFIX = ''
    SLOPES = ('POS', 'NEG')
    POLARITIES = ('NORM', 'INV') 
    VOLTAGE_UNITS = ('VPP', 'VRMS', 'DBM')
    
    # --------------------- MISCELLANEOUS ---------------------
    
    def connect(self):
        super().connect()
        self.disable_output()
        self.beep()
        # self.display_for('instrument connected', duration=1.0)

    def get_version(self):
        ''' Get system SCPI version. '''
        return float(self.query('SYST:VERS?'))

    def get_nchannels(self):
        ''' Get the number of channels available in the instrument '''
        return float(self.query(f'SYST:CHAN:NUM?'))
    
    @abc.abstractmethod
    def beep(self):
        ''' Issue a single beep immediately. '''
        raise NotImplementedError
    
    def check_slope(self, slope):
        if slope not in self.SLOPES:
            raise VisaError(
                f'{slope} is not a valid slope (options are {self.SLOPES})')
    
    def check_polarity(self, pol):
        if pol not in self.POLARITIES:
            raise VisaError(
                f'{[pol]} is not a valid polarity (options are {self.POLARITIES})')
    
    def wait(self):
        self.write('*WAI')
        
    # --------------------- FRONT PANEL & DISPLAY ---------------------

    def lock_front_panel(self):
        self.write('SYST:KLOC:STAT ON')

    def unlock_front_panel(self):
        self.write('SYST:KLOC:STAT OFF')

    @property
    @abc.abstractmethod
    def MAX_LEN_TEXT(self):
        ''' Maximum text length on instrument display '''
        raise NotImplementedError
    
    def display_text(self, text):
        ''' Display text on the instrument screen. '''
        if len(text) > self.MAX_LEN_TEXT:
            raise VisaError(f'Maximum text length ({self.MAX_LEN_TEXT} chars) exceeded.')
        self.write(f'DISP:TEXT "{text}"')

    def erase_text(self):
        ''' Erase text from the instrument screen. '''
        self.write('DISP:TEXT:CLEAR')
    
    def display_for(self, text, duration=1.):
        ''' Display text on the instrument screen for a given duration (in s). '''
        self.chain(lambda: self.display_text(text), self.erase_text, interval=duration)

    # --------------------- ERRORS ---------------------

    def get_last_error(self):
        ''' Get last entry in error queue. '''
        return self.query('SYST:ERR?')
    
    def enable_beep_on_error(self):
        ''' Enable instrument beeping upon error. '''
        self.write('SYST:BEEP:STAT ON')

    def disable_beep_on_error(self):
        ''' Disable instrument beeping upon error. '''
        self.write('SYST:BEEP:STAT OFF')

    def beeps_on_error(self):
        ''' Get whether instrument beeps upon error. '''
        return bool(self.query('SYST:BEEP:STAT?'))

    # --------------------- UNITS ---------------------

    @property
    @abc.abstractmethod
    def ANGLE_UNITS(self):
        raise NotImplementedError

    def check_angle_unit(self, unit):
        if unit not in self.ANGLE_UNITS:
            raise VisaError(
                f'{unit} is not a valid angle unit (options are {self.ANGLE_UNITS})')
    
    @abc.abstractmethod
    def set_angle_unit(self, *args, **kwargs):
        ''' Set the unit used for angles. '''
        raise NotImplementedError

    @abc.abstractmethod        
    def get_angle_unit(self, *args, **kwargs):
        ''' Get the unit used for angles. '''
        raise NotImplementedError
    
    def check_voltage_unit(self, unit):
        if unit not in self.VOLTAGE_UNITS:
            raise VisaError(
                f'{unit} is not a valid voltage unit (options are {self.VOLTAGE_UNITS})')
    
    @abc.abstractmethod
    def set_voltage_unit(self, *args, **kwargs):
        ''' Set the unit used for voltages. '''
        raise NotImplementedError

    @abc.abstractmethod        
    def get_voltage_unit(self, *args, **kwargs):
        ''' Get the unit used for voltages. '''
        raise NotImplementedError

    # --------------------- OUTPUT ---------------------

    @abc.abstractmethod
    def enable_output(self):
        ''' Turn all outputs ON '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def disable_output(self):
        ''' Turn all outputs OFF '''
        raise NotImplementedError

    @abc.abstractmethod    
    def get_output_state(self, *args, **kwargs):
        ''' Get output state '''
        raise NotImplementedError

    # --------------------- BASIC WAVEFORM ---------------------

    @property
    @abc.abstractmethod
    def WAVEFORM_TYPES(self):
        return NotImplementedError

    def check_waveform_type(self, wtype):
        ''' Check that waveform type is available on the instrument '''
        if wtype not in self.WAVEFORM_TYPES:
            raise VisaError(
                f'{wtype} is not a valid waveform (options are {self.WAVEFORM_TYPES})')
    
    @abc.abstractmethod
    def apply_waveform(*args, **kwargs):
        raise NotImplementedError

    def apply_pulse(self, *args, **kwargs):
        self.apply_waveform('PULS', *args, **kwargs)

    def apply_ramp(self, *args, **kwargs):
        self.apply_waveform('RAMP', *args, **kwargs)

    def apply_sine(self, *args, **kwargs):
        self.apply_waveform('SIN', *args, **kwargs)

    def apply_square(self, *args, **kwargs):
        self.apply_waveform('SQU', *args, **kwargs)

    def apply_triangle(self, *args, **kwargs):
        self.apply_waveform('TRI', *args, **kwargs)
    
    def apply_DC(self, *args, **kwargs):
        self.apply_waveform('DC', *args, freq=1, amp=1, **kwargs)

    @abc.abstractmethod
    def set_waveform_type(self, *args, **kwargs):
        ''' Set the nature of the carrier waveform. '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def get_waveform_type(self, *args, **kwargs):
        ''' Get the nature of the carrier waveform. '''
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def FMAX(self):
        raise NotImplementedError

    def check_freq(self, freq):
        if freq > self.FMAX:
            raise VisaError(f'Frequency must be lower than {self.FMAX}')

    @abc.abstractmethod
    def set_waveform_freq(self, *args, **kwargs):
        ''' Set the carrier frequency (Hz). '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def get_waveform_freq(self, *args, **kwargs):
        ''' get the carrier frequency (Hz). '''
        raise NotImplementedError
    
    @property
    @abc.abstractmethod
    def VMAX(self):
        raise NotImplementedError

    def check_amp(self, amp):
        if amp / 2 > self.VMAX:
            raise VisaError(f'VPP exceeds {self.VMAX} V')
    
    @abc.abstractmethod
    def set_waveform_amp(self, *args, **kwargs):
        ''' set the carrier signal amplitude '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_waveform_amp(self, *args, **kwargs):
        ''' Get the carrier signal amplitude '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def check_offset(self, offset, *args, **kwargs):
        ''' Check the waveform voltage offset '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def set_waveform_offset(self, *args, **kwargs):
        ''' Set the waveform voltage offset (V). '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_waveform_offset(self, *args, **kwargs):
        ''' Get the waveform voltage offset (V). '''
        raise NotImplementedError
    
    def check_phase(self, phi):
        ''' Check the waveform phase '''
        if phi < 0. or phi > 360:
            raise VisaError(f'phase out of range : {phi} (must be within [0, 360]deg)')

    @abc.abstractmethod
    def set_waveform_phase(self, *args, **kwargs):
        ''' Set the waveform phase of the specified channel (in degrees) '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def get_waveform_phase(self, *args, **kwargs):
        ''' Get the waveform phase of the specified channel (in degrees) '''
        raise NotImplementedError
    
    def check_duty_cycle(self, DC):
        ''' Check duty cycle '''
        if DC <= 0. or DC >= 100.:
            raise VisaError('Duty cycle out of range (must be within [0-100]%)')
    
    @abc.abstractmethod
    def set_square_duty_cycle(self, *args, **kwargs):
        ''' Set the square wave duty cycle (in %). '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_square_duty_cycle(self, *args, **kwargs):
        ''' Get the square wave duty cycle (in %). '''
        raise NotImplementedError

    # --------------------- ARBITRARY WAVEFORMS ---------------------

    def normalize(self, y, lb=-1, ub=1):
        ''' Normalize AF signal to [-1, 1] interval by linear transformation.

            :param y: AF signal
            :param lb: lower bound
            :param ub: upper bound
        '''
        return (y - y.min()) / np.ptp(y) * (ub - lb) + lb

    # --------------------- BURST ---------------------

    @abc.abstractmethod
    def enable_burst(self, *args, **kwargs):
        ''' Turn burst ON '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def disable_burst(self, *args, **kwargs):
        ''' Turn burst OFF '''
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def BURST_MODES(self):
        raise NotImplementedError
    
    def check_burst_mode(self, mode):
        ''' Check that burst mode is available on the instrument '''
        if mode not in self.BURST_MODES:
            raise VisaError(
                f'{mode} is not a valid burst mode (options are {self.BURST_MODES})')

    @abc.abstractmethod
    def set_burst_mode(self, *args, **kwargs):
        ''' Set burst mode '''
        raise NotImplementedError
    
    @abc.abstractmethod
    def get_burst_mode(self, *args, **kwargs):
        ''' Get burst mode '''
        raise NotImplementedError

    @abc.abstractmethod
    def set_burst_ncycles(self, *args, **kwargs):
        ''' Set number of cycles in burst '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_burst_ncycles(self, *args, **kwargs):
        ''' Get the number of cycles in the burst '''
        raise NotImplementedError

    @abc.abstractmethod
    def set_burst_internal_period(self, T):
        ''' Set burst repetition period (in s). '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_burst_internal_period(self):
        ''' Get burst repetition period (in s). '''
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def MAX_BURST_PERIOD(self):
        raise NotImplementedError

    def check_burst_duration(self, t, T):
        ''' Check the burst duration and return corresponding number of cycles '''
        if t < T:
            if not np.isclose(T - t, 0):
                raise VisaError(
                    f'burst duration ({si_format(t, 2)}s) shorter than stimulus periodicity ({si_format(T, 2)}s)')
        if t > self.MAX_BURST_PERIOD:
            raise VisaError(
                f'burst duration ({t:.2e} s) above max value ({self.MAX_BURST_PERIOD:.2e} s)')
        return int(np.round(t / T))

    @abc.abstractmethod
    def set_burst_duration(self, *args, **kwargs):
        ''' Set burst duration (in s). '''
        raise NotImplementedError

    def get_burst_duration(self, *args, **kwargs):
        ''' Get burst duration (in s). '''
        return self.get_burst_ncycles(*args, **kwargs) / self.get_carrier_freq(*args, **kwargs)

    @abc.abstractmethod
    def get_burst_phase(self, *args, **kwargs):
        ''' Get the burst phase. '''
        raise NotImplementedError

    def check_burst_phase(self, phi):
        unit = self.get_phase_unit()
        half_cycle = 180. if unit == 'DEG' else np.pi
        if phi not in (0., half_cycle):
            raise VisaError(f'burst phase ({phi} {unit}) is not a multiple of the half-cycle')

    @abc.abstractmethod
    def set_burst_phase(self, *args, **kwargs):
        ''' Set the burst phase. '''
        raise NotImplementedError

    @abc.abstractmethod
    def set_burst_gated_polarity(self, *args, **kwargs):
        ''' Set the gate polarity of the gated burst of the specified channel '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_burst_gated_polarity(self, *args, **kwargs):
        ''' Get the gate polarity of the gated burst of the specified channel '''
        raise NotImplementedError
    

    # --------------------- MODULATION ---------------------

    def check_modulation_mode(self, mode):
        ''' Check that modulation mode is available on the instrument '''
        if mode not in self.MOD_MODES:
            raise VisaError(
                f'{mode} is not a valid modulation mode (options are {self.MOD_MODES})')
    
    def check_modulation_source(self, source):
        ''' Check that modulation source is available on the instrument '''
        if source not in self.MOD_SOURCES:
            raise VisaError(
                f'{source} is not a valid modulation source (options are {self.MOD_SOURCES})')
    
    # --------------------- TRIGGER ---------------------

    @property
    @abc.abstractmethod
    def TRIGGER_SOURCES(self):
        raise NotImplementedError

    def check_trigger_source(self, source):
        ''' Check the trigger source '''
        if source not in self.TRIGGER_SOURCES:
            raise VisaError(
                f'{source} is not a valid trigger source (options are {self.TRIGGER_SOURCES})')

    @abc.abstractmethod
    def set_trigger_source(self, *args, **kwargs):
        ''' Select the trigger source. '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_trigger_source(self, *args, **kwargs):
        ''' Read the trigger source. '''
        raise NotImplementedError

    def trigger(self, verbose=True):
        super().trigger()
        if verbose:
            self.display_for('instrument triggered', duration=0.5)

    @abc.abstractmethod
    def wait_for_external_trigger(self, *args, **kwargs):
        ''' Wait for an external trigger '''
        raise NotImplementedError

    @abc.abstractmethod
    def wait_for_manual_trigger(self, *args, **kwargs):
        ''' Wait for a manual/programmatic trigger '''
        raise NotImplementedError

    # --------------------- TRIGGER OUTPUT ---------------------
    
    @abc.abstractmethod
    def enable_trigger_output(self):
        ''' Enable trigger output '''
        raise NotImplementedError

    @abc.abstractmethod
    def disable_trigger_output(self):
        ''' Disable trigger output '''
        raise NotImplementedError

    @abc.abstractmethod
    def set_trigger_output_slope(self, slope):
        ''' Set the trigger output slope type '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_trigger_output_slope(self):
        ''' Read the trigger output slope type. '''
        raise NotImplementedError

    def disconnect(self):
        self.unlock_front_panel()
        super().disconnect()


# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Email: theo.lemaire@epfl.ch
# @Date:   2017-06-13 09:40:02
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-08-08 16:18:18

import os
from setuptools import setup

readme_file = 'README.md'
req_file = 'requirements.txt'


def readme():
    with open(readme_file, encoding='utf8') as f:
        return f.read()

def getRequirements():
    with open(req_file, encoding='utf8') as f:
        return f.readlines()

def getFiles(path):
    return [f'{path}/{x}' for x in os.listdir(path)]

setup(
    name='instrulink',
    version='1.0.2',
    description='Python package to interface diverse laboratory instruments',
    long_description_content_type='text/markdown',
    long_description=readme(),
    url='https://github.com/tjjlemaire/instrulink',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Physics'
    ],
    keywords=('laboratory instrument interface python'),
    author='Theo Lemaire',
    author_email='theo.lemaire1@gmail.com',
    license='MIT',
    packages=['instrulink'],
    scripts=getFiles('scripts'),
    install_requires=getRequirements(),
    zip_safe=False
)


# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-04-27 18:16:34
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-11 16:03:27

import serial
import struct
import time
import numpy as np

from .logger import logger
from .utils import is_within


class SutterError(Exception):
    ''' Custom exception class for Sutter instrument '''
    pass


class SutterMP285A:
    '''' Interface to communicate with Sutter Manipulator 285 '''

    PORT = 'COM2'  # Default port
    BAUDRATE = 9600  # Default baud rate
    INTERCOMMAND_DELAY = 2e-3  # internal delay between commands (s)
    ENC = 'utf-8'
    CR = bytes('\r', ENC)
    XYZ_FMT = 'lll'
    XYZ_LEN = 12  # length (in bytes) of XYZ position
    XYZ_BOUNDS = (-12500, 12500)  # Bounds for XYZ coordinates (um)
    XYZ_TOL = 0.5  # XYZ tolerance (um)
    TREL_MAX = 0.7  # maximal relative move duration w.r.t. timeout duration
    TREL_WARN = 1.8  # critical relative move duration w.r.t. estimate above which to throw a warning
    LOW_RES_SPEED_BOUNDS = (0, 3000)  # Coarse mode travel speed bounds (um/s)
    HIGH_RES_SPEED_BOUNDS = (0, 1310)  # Fine mode travel speed bounds (um/s)
    STATUS_LEN = 32  # length (in bytes) of status information
    STATUS_FIELDS = {
        'FLAGS': 'B',  # Program flags
        'UDIRX': 'B',  # User-defined values for motor axis X
        'UDIRY': 'B',  # User-defined values for motor axis Y
        'UDIRZ': 'B',  # User-defined values for motor axis Z
        'ROE_VARI': 'H',  # u-steps per ROE click
        'UOFFSET': 'H',  # User-defined period start value
        'URANGE': 'H',  # User-defined period range
        'PULSE': 'H',  # Number of u-steps per pulse
        'USPEED': 'H',  # Adjusted pulse speed (u-steps/s)
        'INDEVICE': 'B',  # Input device type
        'FLAGS_2': 'B',  # Program flags
        'JUMPSPD': 'H',  # "Jump to max at" speed
        'HIGHSPD': 'H',  # "Jumped to" speed
        'DEAD': 'H',  # Dead zone, not saved
        'WATCH_DOG': 'H',  # programmer's function
        'STEP_DIV': 'H',  # Divisor yields (u-steps/um)
        'STEP_MUL': 'H',  # Multiplier yields (um/u-steps)
        'XSPEED': 'H',  # Remote speed (um/s) & res. (bit 15)
        'VERSION': 'H'  # Firmware version
    }
    FLAGS_FIELDS = {
        'SETUP': [0, 1, 2, 3],  # Currently loaded setup number.
        'ROE_DIR': 4,  # ROE direction
        'REL_ABS_F': 5,  # Display origin
        'MODE_F': 6,  # Manual mode
        'STORE_F': 7  # Setup condition
    }
    FLAGS_2_FIELDS = [
        'LOOP_MODE',  # Program loops
        'LEARN_MODE',  # Learn mode status
        'STEP_MODE',  # Movement resolution
        'SW2_MODE',  # Enable joy side button
        'SW1_MODE',  # Enable FSR/Joystick
        'SW3_MODE',  # Enable ROE switch
        'SW4_MODE',  # Enable SW 4&5
        'REVERSE_IT',  # Reverse program
    ]
    BIT15 = 2**15  # Bit 15 position value

    def __init__(self, timeout=10., lock=False):
        '''
        Initialization

        :param timeout: commands timeout (s)
        :param lock: whether to lock the instrument during motion (not implemented)
        '''
        self.connect()
        self.timeout = timeout
        self.set_absolute_mode()
        self.update_status()
        self.set_resolution(1)
        self.set_velocity(self.HIGH_RES_SPEED_BOUNDS[1])
    
    def __repr__(self):
        ''' String representation '''
        return f'{self.__class__.__name__}({self.instrument_handle.port})'
    
    def connect(self):
        ''' Attempt serial connection to controller '''
        try:
            self.instrument_handle = serial.Serial(
                port=self.PORT,
                baudrate=self.BAUDRATE,
                bytesize=serial.EIGHTBITS,
                parity=serial.PARITY_NONE,
                stopbits=serial.STOPBITS_ONE)
            print(f'{f" {repr(self)} ":-^100}')
            
        except serial.SerialException:
            self.instrument_handler = None
            raise SutterError(
                f'No connection to {self.__class__.__name__} could be established')
    
    def disconnect(self):
        ''' Disconnect from controller '''
        logger.info(f'disconnecting from {self}')
        self.instrument_handle.close()
        self.instrument_handle = None

    @property
    def is_connected(self):
        ''' Whether the instrument is connected '''
        return self.instrument_handle is not None
    
    @property
    def timeout(self):
        ''' Timeout (s) '''
        return self.instrument_handle.timeout
    
    @timeout.setter
    def timeout(self, value):
        ''' Set timeout (s) '''
        self.instrument_handle.timeout = value

    # def __del__(self):
    #     logger.info(f'releasing {repr(self)} resource')
    #     self.instrument_handle.close() 

    def get_name(self):
        ''' Get controller name '''
        return self.__class__.__name__
    
    def write(self, cmd, convert_to_bytes=True):
        ''' Write a command into the serial instrument '''
        logger.debug(f'WRITE: {cmd}')
        if convert_to_bytes:
            cmd = bytes(cmd, self.ENC)
        self.instrument_handle.write(cmd + self.CR)
        time.sleep(self.INTERCOMMAND_DELAY)
    
    def read(self, n):
        '''
        Read a specific number of bytes from the instrument response
        
        :param n: number of bytes to read
        :return: response bytes sequence
        '''
        return self.instrument_handle.read(n)
    
    def read_until_CR(self, lexp=None):
        '''
        Read instrument response until carriage return character
        
        :param lexp (optional): expected response length (number of bytes)
        :return: response bytes sequence
        '''
        out = self.instrument_handle.read_until(self.CR)[:-1]
        if lexp is not None:
            if len(out) != lexp:
                raise SutterError(f'expected {lexp}-bytes long response, got {len(out)} bytes')
        return out
    
    def write_and_check(self, cmd, **kwargs):
        '''
        Execute "set" type command and check for correct execution
        
        :param cmd: command string
        '''
        tstart = time.time()  # start timer
        self.write(cmd, **kwargs)  # send "set" command to controller
        out = self.read(1)  # read output
        tend = time.time()  # stop timer
        if out != self.CR:  # check that output is expected
            raise SutterError(
                f'command {cmd} did not complete before timeout ({self.timeout} s)')
        logger.debug(f'completed in {tend - tstart:.2f} s')
    
    @property
    def status_fmt(self):
        ''' Format of the status fields for unpacking '''
        return ''.join(self.STATUS_FIELDS.values())
    
    def get_status(self):
        '''
        Query the status of the controller.
        
        :return: status data dictionary
        '''
        # Query status data
        self.write('s')
        # Read status data block
        status_data_block = self.read_until_CR(lexp=self.STATUS_LEN)
        # Unpack block and format into dictionary
        status_data = struct.unpack(self.status_fmt, status_data_block)
        status_data = dict(zip(self.STATUS_FIELDS.keys(), status_data))
        # Post-process specific fields
        status_data['VERSION'] /= 100  # divide version by 100
        status_data['STEP_MUL'] = 10000 / status_data['STEP_MUL']  # convert step_mul to u-steps/um
        status_data['STEP_DIV'] /= 10000  # convert step_div to um/u-steps
        if status_data['XSPEED'] >= self.BIT15:  # if high resolution
            status_data['SPEED_RES'] = 1  # set resolution to high
            status_data['XSPEED'] = status_data['XSPEED'] - self.BIT15  # correct speed
        else:
            status_data['SPEED_RES'] = 0  # set resolution to low
        # Log dict
        status_str = '\n'.join([f'  - {k}: {v}' for k, v in status_data.items()])
        logger.debug(f'status data:\n{status_str}')
        return status_data

    def update_status(self):
        ''' Update controller status '''
        self.status_data = self.get_status()
    
    @property
    def usteps_per_um(self):
        ''' Get u-steps per um '''
        return self.status_data['STEP_MUL']

    def encode_position(self, pos, prefix=None):
        '''
        Encode an XYZ position into bytes
        
        :param pos: position vector (in um)
        '''
        # Convert to u-steps and integers
        pos_usteps = (pos * self.usteps_per_um).astype(int)
        # Pack and return
        bpos = struct.pack(self.XYZ_FMT, *pos_usteps)
        if prefix is not None:
            bpos = bytes(prefix, self.ENC) + bpos
        return bpos

    def decode_position(self):
        '''
        Read and decode an XYZ position
        
        :return: position vector (in um)
        '''
        # Read binary position from controller
        bpos = self.read(self.XYZ_LEN + 1)[:-1]
        # bpos = self.read_until_CR(lexp=self.XYZ_LEN)
        # Decode u-steps position
        pos_usteps = np.array(struct.unpack(self.XYZ_FMT, bpos))
        # Convert to um and return
        return pos_usteps / self.usteps_per_um  # um
    
    def pos_str(self, pos):
        ''' Return a string representation of a position vector '''
        pos_str = ', '.join([f'{x:.2f}' for x in pos])
        return f'[{pos_str}] um'
    
    def get_position(self, verbose=False):
        '''
        Get controller XYZ position (in um)
        
        :return: 1D (X, Y, Z) array (um)
        '''
        # Send commend to get position
        self.write('c')
        # Read position from controller
        pos = self.decode_position()
        logfunc = logger.info if verbose else logger.debug
        logfunc(f'stage position: {self.pos_str(pos)}')
        return pos
    
    def check_coordinate(self, k, v):
        ''' Check that coordinate is within bounds '''
        if not is_within(v, self.XYZ_BOUNDS):
            raise SutterError(f'target {k} coordinate {v:.3f} um is out of bounds {self.XYZ_BOUNDS}')
    
    def check_coordinates(self, cdict):
        ''' Check a dictionary of coordinates '''
        for k, v in cdict.items():
            self.check_coordinate(k, v)
    
    def set_position(self, pos, silent=False):
        '''
        Moves the three axes to specified location (in um).
        
        :param pos: 1D (X, Y, Z) array (um)
        :param silent: whether to log the move
        '''
        # Check validity of target position
        pos = np.asarray(pos)
        if pos.size != 3:
            raise SutterError('input position must be of length 3')
        self.check_coordinates(dict(zip('XYZ', pos)))
        # Compute delta to target
        current_pos = self.get_position()
        delta = pos - current_pos
        dtot = np.linalg.norm(delta)
        # If total distance is zero (no delta), return
        if dtot == 0:
            logger.warning('already at position!')
            return
        # Check if velocity will allow to move there before timeout
        v = self.get_velocity()  # um/s
        tmove_est = dtot / v # s
        # If not, check required speed to achieve it in percentage of timeout
        tcr = self.TREL_MAX * self.timeout
        vreq = None
        if tmove_est > self.timeout:
            vreq = int(np.round(dtot / tcr))
            logger.warning(
                f'increasing velocity temporarily to {vreq} um/s to cover {dtot:.2f} um within {self.TREL_MAX * 1e2:.0f} % of {self.timeout} s timeout')
            self.set_velocity(vreq)
        if not silent:
            logger.info(f'moving to position: {self.pos_str(pos)} (delta = {self.pos_str(delta)})')
        bpos = self.encode_position(pos, prefix='m')  # encode position to bytes
        tmove = time.perf_counter()
        self.write_and_check(bpos, convert_to_bytes=False)  # send position to controller
        tmove = time.perf_counter() - tmove
        # Check that target position has been reached
        new_pos = self.get_position()
        new_delta = pos - new_pos
        if np.sum(np.abs(new_delta)) > self.XYZ_TOL:
            raise SutterError(f'could not reach target position {self.pos_str(pos)} (value = {self.pos_str(new_pos)})')
        if tmove > max(50e-3, self.TREL_WARN * tmove_est):
            logger.warning(f'{delta} move took {tmove * 1e3:.3f} ms ({tmove / tmove_est:.2f} times expected time)')
        if vreq is not None:
            logger.warning(f'resetting velocity to {v} um/s')
            self.set_velocity(v)
    
    def move_to_origin(self):
        ''' Move to origin '''
        self.set_position([0., 0., 0.])
        logger.info('moved to origin!')
    
    def translate(self, v, **kwargs):
        '''
        Translate by some vector
        
        :param v: XYZ translation vector (um) '''
        pos = self.get_position()
        self.set_position(pos + np.asarray(v), **kwargs)
    
    def get_velocity(self, verbose=False):
        ''' Get controller motion velocity (um/s) '''
        v = self.status_data['XSPEED']
        logfunc = logger.info if verbose else logger.debug
        logfunc(f'velocity = {v} um/s')
        return v
    
    def get_resolution(self, verbose=False):
        ''' Get controller motion resolution (0 or 1) '''
        res = self.status_data['SPEED_RES']
        logfunc = logger.info if verbose else logger.debug
        logfunc(f'motion mode: {self.res_str(res)}')
        return res
    
    def encode_velocity_and_resolution(self, v, res):
        '''
        Encode velocity and resolution
        
        :param v: target velocity (um/s)
        :param res: resolution (0 or 1)
        '''
        # Encode speed and resolution into single integer
        vres = int(v) + res * self.BIT15
        # Convert to unsigned short (2 bytes)
        bvres = struct.pack('H', vres)
        # Send command and check response
        self.write_and_check(bytes('V', self.ENC) + bvres, convert_to_bytes=False)
        self.update_status()

    def get_vbounds(self, res):
        '''
        Determine velocity bounds depending on resolution
        
        :param res: resolution (0 or 1)
        :return: velocity bounds
        '''
        if res:
            return self.HIGH_RES_SPEED_BOUNDS
        else:
            return self.LOW_RES_SPEED_BOUNDS
        
    def set_velocity(self, v):
        ''' 
        Set controller motion velocity (um/s)
        
        :param v: target velocity (um/s)
        '''
        # Check that velocity is within bounds
        res = self.get_resolution()
        vbounds = self.get_vbounds(res)
        if not is_within(v, vbounds):
            raise SutterError(
                f'velocity value ({v} um/s) is out of specific bounds ({vbounds} um/s)')
        logger.info(f'setting velocity to {v} um/s')
        self.encode_velocity_and_resolution(v, res)
        vout = self.get_velocity()
        if vout != v:
            raise SutterError(f'could not set motion speed to {v} um/s (value = {vout} um/s)')
    
    def res_str(self, res):
        ''' Return a string representation of controller movement resolution '''
        resmode = {0: 'low', 1: 'high'}[res]
        return f'{resmode}-resolution'
    
    def set_resolution(self, res):
        ''' Set controller motion resolution (0 for low or 1 for high) '''
        logger.info(f'setting motion mode to {self.res_str(res)}')
        self.encode_velocity_and_resolution(self.get_velocity(), res)
        resout = self.get_resolution()
        if resout != res:
            raise SutterError(f'could not set motion mode to {self.res_str(res)} (value = {self.res_str(resout)})')

    def refresh_panel(self):
        ''' Refresh XYZ info on the front panel '''
        self.write_and_check('n')
    
    def set_origin(self):
        ''' Set origin of the coordinate system to the current position '''
        self.write_and_check('o')
        logger.info('origin reset!')
    
    def reset(self):
        ''' Reset controller '''
        self.write('r')  # controller does not reply
    
    def set_absolute_mode(self):
        ''' Set the movement mode to absolute '''
        logger.info('setting absolute movement mode')
        self.write_and_check('a')
    
    def set_relative_mode(self):
        ''' Set the movement mode to relative '''
        logger.info('setting relative movement mode')
        self.write_and_check('b')



# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-03-15 15:44:20
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-08-07 14:54:46

''' Initiate test sequence with Rigol waveform generator. '''

import logging
import matplotlib.pyplot as plt
import seaborn as sns
from tqdm import tqdm

from instrulink.rigol_dg1022z import RigolDG1022Z
from instrulink import logger
from instrulink.wf_utils import *

# Set logger level
logger.setLevel(logging.INFO)

# Waveform parameters
Fdrive = 2.1e6  # carrier frequency (Hz)
Vpp = 2  # signal amplitude (Vpp)
tstim = 50e-3  # burst duration (s)
PRF = 100.  # burst internal PRF (Hz)
DC = 50.  # burst internal duty cycle (%)
npts = RigolDG1022Z.ARB_WF_MAXNPTS_PER_PACKET  # nominal waveform envelope vector size
tramp_bounds = (0, 5e-3)  # ramp up time range for smoothed waveforms (s)
tramps = np.linspace(*tramp_bounds, 20)  # ramp up times (s)

# Determine which ramps to actually plot
nplt = 3
if nplt < tramps.size:
    iplt = np.linspace(0, tramps.size - 1, nplt).astype(int)
else:
    iplt = np.arange(tramps.size)

# Determine color map
cmap = sns.color_palette('crest', as_cmap=True)

# Initialize vector to store log-frequency spectrum value (in dB) at PRF 
# for each ramp time
PRFdb = np.zeros(tramps.size)

# Create figure for spectrum plot
fig, axes = plt.subplots(1, 2, figsize=(8, 4))

# Loop over ramp up times
logger.info('looping through ramp up times and generating waveforms')
for i, tramp in tqdm(enumerate(tramps)):
    # Generate label
    label = f'tramp = {tramp * 1e3:.2f} ms'

    # Generate waveform
    t, y = get_DC_smoothed_pulse_envelope(npts, PRF, DC, tramp=tramp)

    # # Plot waveform
    # plot_smoothed_waveform(
    #     t, y, Fdrive, ax=axes[0], nreps=5, label=label, mark_regs=False, 
    #     title='waveforms comparison')
    
    # Plot spectrum
    _, PRFdb[i] = plot_waveform_spectrum(
        t, y, Fdrive, ax=axes[0], nreps=20, 
        plot=i in iplt, label=label, 
        mark_freqs=i == len(tramps) - 1, title='spectra comparison',
        get_PRF_val=True, color=cmap(i / (len(tramps) - 1)))

# Plot log-spectrum value at PRF as a function of ramp up time
ax = axes[1]
sns.despine(ax=ax)
ax.set_title(f'spectrum attenuation at PRF = {PRF:.2f} Hz')
ax.plot(tramps * 1e3, PRFdb, c='k', marker='o')
ax.set_xlabel('ramp up time (ms)')
ax.set_ylabel('log-spectrum value at PRF (dB)')

# Adjust figure layout and render
fig.tight_layout()
plt.show()



# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-08-15 09:29:37
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-08-07 14:50:28

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from scipy.signal import welch

from .logger import logger
from .si_utils import si_prefixes, si_format


def sigmoid_ramp(t, tramp=1, t0=0, dy=0.01):
    ''' 
    Sigmoidal ramp-up function with a specific ramp time and half-ramp point
    
    :param t: time vector
    :param tramp: ramp time
    :param t0: half-ramp point
    :param dy: vertical offset for ramp time (default = 0.01)
    :return: horizontally and vertically scaled sigmoid ramp-up vector
    '''
    # Get sigmoid scaling factor from analytical solution of sigmoid function
    if dy <= 0 or dy >= 1:
        raise ValueError('dy must be between 0 and 1')
    k = 2 * np.log((1 - dy) / dy) / tramp
    # Get sigmoid vector
    y = 1 / (1 + np.exp(-k * (t - t0)))
    # Rescale sigmoid vector to [0 - 1] range
    y = (y - dy) / (1 - 2 * dy)
    # Return sigmoid vector
    return y


def sine_ramp(t, tramp=1, t0=0):
    ''' 
    Sine ramp-up function with a specific ramp time and half-ramp point

    :param t: time vector
    :param tramp: ramp time
    :param t0: half-ramp point
    :return: horizontally and vertically scaled sinusoidal ramp-up vector
    '''
    return (np.sin(2 * np.pi * (t - t0) / (2 * tramp)) + 1) / 2


def halfsine_ramp(t, tramp=1, t0=0):
    ''' 
    Half-sine ramp-up function with a specific ramp time and half-ramp point

    :param t: time vector
    :param tramp: ramp time
    :param t0: half-ramp point
    :return: horizontally and vertically scaled sinusoidal ramp-up vector
    '''
    return np.sin(2 * np.pi * (t - t0) / (4 * tramp) + np.pi / 4)


def get_smooth_ramp(n, tramp=1, t0=0, kind='sine', direction='up', **kwargs):
    '''
    Get a smooth ramp vector to transition from 0 to 1 within a specific ramp time and half-ramp point
    
    :param n: number of points in the ramp vector
    :param tramp: ramp time
    :param t0: half-ramp point
    :param kind: ramp-up function type ('sigmoid' or 'sine')
    :param direction: ramp direction ('up' or 'down')
    :return: time and sigmoid vectors
    '''
    if tramp <= 0:
        raise ValueError('tramp must be positive')
    # Get time vector
    t = np.linspace(0, tramp, n) + t0 - tramp / 2
    # Get ramp-up vector
    if kind == 'sigmoid':
        y = sigmoid_ramp(t, t0=t0, tramp=tramp, **kwargs)
    elif kind == 'sine':
        y = sine_ramp(t, t0=t0, tramp=tramp)
    elif kind == 'halfsine':
        y = halfsine_ramp(t, t0=t0, tramp=tramp)
    else:
        raise ValueError(f'Invalid ramping function type: {kind}')
    # Invert ramping vector if "down" direction is specified
    if direction == 'down':
        y = y[::-1]
    # Return time and ramp vectors
    return t, y


def get_smoothed_pulse_envelope(n, tramp, thigh, tlow=0, plot=None, Fdrive=None, nreps=1, **kwargs):
    '''
    Get a smoothed rectangular pulse envelope with a specific ramp-up (and down) time,
    plateau duration and base duration.
    
    :param n: number of points in envelope vector
    :param tramp: pulse ramp-up (and down) time
    :param thigh: pulse plateau duration
    :param tbase: post-pulse baseline duration
    :param plot: flag indicating whether/what to plot:
        - None: nothing (default)
        - "wf": the waveform envelope
        - "fft": the waveform frequency spectrum
        - "all": both envelope and spectrum
    :param Fdrive: carrier frequency (Hz) used for plotting purposes.
        Defaults to 25 / thigh if not specified.
    :param nreps: number of nominal envelope repetitions for plotting (default = 1)
    :return: time and envelope vectors
    '''
    # Compute characteristic times and total duration 
    tonset = 0
    toffset = tramp + thigh
    ttot = 2 * tramp + thigh + tlow

    # Derive number of points in ramp-up (and down) phases
    nramp = int(np.round(n * tramp / ttot)) + 1
    dt = ttot / (n - 1)

    # Get data points for rising and falling edges
    try:
        tr, yr = get_smooth_ramp(
            nramp, tramp=tramp, t0=tonset + tramp / 2, direction='up', **kwargs)
        tf, yf = get_smooth_ramp(
            nramp, tramp=tramp, t0=toffset + tramp / 2, direction='down', **kwargs)
    except ValueError as e:
        tr, yr = np.array([0.]), np.array([0.])
        tf, yf = np.array([0.]), np.array([1.])
    
    # Check that ramp time is long enough to be resolved by the requested time step
    if tr.size > 1:
        dt = np.diff(tr)[0]
    else:
        logger.warning(
            f'ramping time ({si_format(tramp, 2)}s) is too short to be resolved by the requested time step ({si_format(dt, 2)}s) -> switching to discrete transition')
    
    # Get data points for pulse "high" plateau
    th = np.arange(tr[-1] + dt, toffset - dt / 2, dt)
    yh = np.ones(th.size)

    # If plateau is empty, remove first ramp-down point
    if th.size == 0:
        tf = tf[1:]
        yf = yf[1:]

    # Get data points for pulse "low" baseline
    tl = np.arange(tf[-1] + dt, ttot + dt / 2, dt)
    yl = np.zeros(tl.size)

    # Concatenate data points
    tenv = np.hstack([tr, th, tf, tl])
    yenv = np.hstack([yr, yh, yf, yl])

    # Interpolate vectors to ensure that they match the requested number of points
    tnew = np.linspace(tenv[0], tenv[-1], n)
    yenv = np.interp(tnew, tenv, yenv)
    tenv = tnew

    # If plot requested
    if plot is not None:
        # Check validity of plot option
        pltchoices = ('wf', 'fft', 'all')
        if plot not in pltchoices:
            raise ValueError(f'Invalid plot option: "{plot}". Valid options are: {pltchoices}')

        # Determine what to plot and generate figure and axes
        pltwf = plot in ('wf', 'all')
        pltfft = plot in ('fft', 'all')
        naxes = int(pltwf) + int(pltfft)
        _, axes = plt.subplots(naxes, 1, figsize=(6, 3 * naxes))
        axes = np.atleast_1d(axes)
        iax = 0

        # Compute carrier frequency if not specified
        if Fdrive is None:
            Fdrive = 25 / thigh  # Hz
        
        # If requested, plot waveform and its envelope
        if pltwf:
            plot_smoothed_waveform(tenv, yenv, Fdrive, ax=axes[iax], nreps=nreps)
            iax += 1
        
        # If requested, plot waveform spectrum
        if pltfft:
            plot_waveform_spectrum(tenv, yenv, Fdrive, ax=axes[iax], nreps=nreps)

    # Return time and envelope vectors
    return tenv, yenv


def get_DC_smoothed_pulse_envelope(n, PRF, DC, tramp=0, **kwargs):
    '''
    Wrapper around get_smoothed_pulse_envelope to get a smoothed rectangular 
    pulse envelope based on a specific pulsing rate and duty cycle.

    :param n: number of points in envelope vector
    :param PRF: pulse repetition frequency (Hz)
    :param DC: duty cycle (%)
    :param tramp: ramp-up (and down) time (s)
    '''
    # Normalize duty cycle to [0 - 1] range
    DC /= 100  # (0 - 1)

    # Compute ON, OFF and total envelope duration
    PRI = 1 / PRF  # s
    ton = DC * PRI  # s
    toff = PRI - ton  # s
    
    # Compute plateau duration and post-pulse baseline duration
    thigh = ton - tramp  # ON time minus half ramp-up and half ramp-down time (ms)
    tlow = toff - tramp  # OFF time minus ramp-down time (ms)

    # Check validity of resulting high and low times
    if thigh < 0:
        raise ValueError(f'ON duration ({si_format(ton, 2)}s) is too short for the specified ramp time ({si_format(tramp, 2)}s)')
    if tlow < 0:
        raise ValueError(f'OFF duration ({si_format(toff, 2)} ms) is too short for the specified ramp time ({si_format(tramp, 2)}s)')

    # Check that total envelope duration does not exceed the PRI
    tenv = thigh + tlow + 2 * tramp
    if tenv > PRI + 1e-12:
        raise ValueError(f'smoothed pulse duration ({si_format(tenv, 5)}s) exceeds pulse repetition interval ({si_format(PRI, 2)}s)')

    # Assemble and return smoothed pulse envelope
    return get_smoothed_pulse_envelope(n, tramp, thigh, tlow=tlow, **kwargs)


def get_full_waveform(t, yenv, Fdrive, npc=25, nreps=1):
    ''' 
    Get a full waveform from a time vector, an envelope vector and a carrier frequency.

    :param t: time vector (s)
    :param yenv: envelope vector
    :param Fdrive: carrier frequency (Hz)
    :param npc: number of points per cycle (default = 25)
    :param nreps: number of nominal waveform repetitions (default = 1)
    :return: dense time, waveform and envelope vectors
    '''
    # Determine sampling frequency and number of points in full waveform
    fs = Fdrive * npc  # Hz
    npts = int(np.round(t[-1] * fs))
    # Generate dense time vector 
    tdense = np.linspace(0, t[-1], npts)  # s
    # Interpolate envelope vector
    yenvdense = np.interp(tdense, t, yenv)
    # Generate carrier vector
    ycarrier = np.sin(2 * np.pi * Fdrive * tdense)
    # Generate waveform vector
    ydense = yenvdense * ycarrier
    # Repeat vectors if needed
    if nreps > 1:
        tdenseexp = tdense.copy()
        ydenseexp = ydense.copy()
        yenvdenseexp = yenvdense.copy()
        for _ in range(2, nreps + 1):
            tdenseexp = np.hstack([tdenseexp, tdenseexp[-1] + tdense[1:]])
            ydenseexp = np.hstack([ydenseexp, ydense[1:]])
            yenvdenseexp = np.hstack([yenvdenseexp, yenvdense[1:]])
        tdense = tdenseexp
        ydense = ydenseexp
        yenvdense = yenvdenseexp
    # Return dense time, waveform and envelope vectors
    return tdense, ydense, yenvdense


def get_power_spectrum(t, y):
    '''
    Get the power spectrum of a waveform.

    :param t: time vector (s)
    :param y: waveform vector
    :return: frequency (Hz) and power vectors
    '''
    dt = np.diff(t)[0]  # s
    freqs = np.fft.rfftfreq(y.size, dt)  # Hz
    ps = np.abs(np.fft.rfft(y))**2  # power
    return freqs, ps


def get_continous_intervals(t, refdt):
    '''
    Identify intervals of continuity in time vector.

    :param t: time vector (s)
    :param refdt: expected time step in continous time vector (s)
    :return: list of (start, end) pairs for each identified continuous time interval
    '''
    # Check that time vector contains at least 2 elements
    if t.size < 2:
        raise ValueError('Time vector must contain at least 2 elements')
    # Check that time contains at least 1 continous segment
    if not np.isclose(np.diff(t), refdt).any():
        raise ValueError('No continuity detected in time vector')
    
    # Identify discontinuities
    dt2 = np.diff(t, 2)
    iscontdt = np.isclose(dt2, 0)
    idiscont = np.where(~iscontdt)[0]

    # Identify index boundaries of continuous intervals
    istarts, iends = [0], []
    for i1, i2 in zip(idiscont[::2], idiscont[1::2]):
        if i2 != i1 + 1:
            raise ValueError(f'invalid discontinuity indices: {i1}, {i2}')
        iends.append(i2)
        istarts.append(i2 + 1)
    iends.append(t.size - 1)

    # Extract corresponding start and end times for each continuity region
    tstarts, tends = t[istarts], t[iends]

    # Return list of (start time, end time) pairs
    return np.array(list(zip(tstarts, tends)))


def get_bounds_per_region(t, y):
    ''' 
    Identify ramp-up, plateau, ramp-down, and baseline regions in waveform vector
    
    :param t: time vector (s)
    :param y: waveform vector
    :return: dictionary of (start, end) pairs for each identified region
    '''
    # Check that time step is consistent
    dt = np.unique(np.round(np.diff(t), 10))
    if dt.size > 1:
        raise ValueError(f'Inconsistent time step detected: {dt}')
    dt = dt[0]

    # Identify ramp-up, plateau, ramp-down, and baseline regions
    tregions = {
        'ramp-up': t[np.where(np.diff(y) > 0)],
        'plateau': t[np.where(y == 1)],
        'ramp-down': t[np.where(np.diff(y) < 0)],
        'baseline': t[np.where(y == 0)][1:],
    }

    # Identify time bounds for each valid region
    tbounds = {}
    for key, tvec in tregions.items():
        try:
            tbounds[key] = get_continous_intervals(tvec, dt)
        except ValueError as e:
            logger.warning(f'{key}: {e}')
    
    # Return dictionary of time bounds for each identified region
    return tbounds


def plot_smoothed_waveform(tenv, yenv, Fdrive, unit='ms', ax=None, title=None, 
                           label=None, mark_regs=True, **kwargs):
    ''' 
    Plot a smoothed waveform from a time vector, an envelope vector and a carrier frequency.

    :param t: time vector (s)
    :param yenv: envelope vector
    :param Fdrive: carrier frequency (Hz)
    :param unit: time unit for plotting (default = 'ms')
    :param ax: axis handle (default = None)
    :param title: title for the plot (default = None)
    :param label: label for the plotted waveform (default = None)
    :param mark_regs: flag indicating whether to mark identified regions (default = True)
    :return: figure handle
    '''
    # Get figure and axis handles
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()
    sns.despine(ax=ax)

    # Get dense time, waveform and envelope vectors
    tdense, ydense, yenvdense = get_full_waveform(tenv, yenv, Fdrive, **kwargs)
    tfactor = si_prefixes[unit[:-1]]
    tdense /= tfactor

    # Plot waveform and its envelope
    lh, *_ = ax.plot(tdense, yenvdense, label=label if label is not None else 'envelope')
    ax.plot(tdense, ydense, c=lh.get_color(), alpha=0.5, label='waveform' if label is None else None)
    ax.set_xlabel(f'time ({unit})')
    ax.set_ylabel('amplitude')
    ax.axhline(0, c='k', ls='--')

    # Identify ramp-up, plateau, ramp-down, and baseline regions
    tbounds = get_bounds_per_region(tdense, yenvdense)
    
    # Get number of pulses
    npulses = max(v.shape[0] for v in tbounds.values())

    # Generate title with breakdown of characteristic waveform phases if not provided
    if title is None:
        l = []
        if 'ramp-up' in tbounds:
            l.append(f'tramp = {np.diff(tbounds["ramp-up"][0])[0]:.2f} {unit}')
        elif 'ramp-down' in tbounds:
            l.append(f'tramp = {np.diff(tbounds["ramp-down"][0])[0]:.2f} {unit}')
        if 'plateau' in tbounds:
            l.append(f'tplateau = {np.diff(tbounds["plateau"][0])[0]:.2f} {unit}')
        if 'baseline' in tbounds:
            l.append(f'tbase = {np.diff(tbounds["baseline"][0])[0]:.2f} {unit}')
        if npulses > 1:
            l.append(f'{npulses} pulses')
        title = ', '.join(l)

    # Add title 
    ax.set_title(title)

    # Mark identified regions, if specified
    if mark_regs:
        colors_per_reg = {
            'ramp-up': 'g',
            'plateau': 'k',
            'ramp-down': 'r',
            'baseline': 'dimgray'
        }
        for reg, tvec in tbounds.items():
            for i, (tstart, tend) in enumerate(tvec):
                ax.axvspan(
                    tstart, tend, color=colors_per_reg[reg], alpha=0.1, 
                    label=reg if i == 0 else None)

    # Add legend
    ax.legend()

    # Return figure handle
    return fig


def plot_waveform_spectrum(tenv, yenv, Fdrive, ax=None, label=None, mark_freqs=True, 
                           title=None, get_PRF_val=False, color=None, plot=True, **kwargs):
    ''' 
    Plot a smoothed waveform from a time vector, an envelope vector and a carrier frequency.

    :param t: time vector (s)
    :param yenv: envelope vector
    :param Fdrive: carrier frequency (Hz)
    :param ax: axis handle (default = None)
    :param label: label for the plotted spectrum (default = None)
    :param mark_freqs: flag indicating whether to mark carrier and PRF frequencies (default = True)
    :param title: title for the plot (default = None)
    :param get_PRF_val: flag indicating whether to compute and return the log-spectrum value (in dB) at the waveform PRF (default = False)
    :return: figure handle (and log-spectrum value if requested)
    '''
    # Get figure and axis handles
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()
    sns.despine(ax=ax)

    # Generate label if not provided
    if label is None:
        label = 'spectrum'
    
    # Add title
    ax.set_title('waveform spectrum' if title is None else title)

    # Get dense time, waveform and envelope vectors
    tdense, ydense, yenvdense = get_full_waveform(tenv, yenv, Fdrive, **kwargs)

    # Extract and waveform frequency spectrum
    freqs, ps = get_power_spectrum(tdense, ydense)
    ps_decibel = 10 * np.log10(ps / ps.max())
    if plot:
        ax.plot(freqs, ps_decibel, label=label, c=color)
    ax.set_xlabel('frequency (Hz)')
    ax.set_ylabel('relative power (dB)')
    ax.set_xscale('log')

    # Get number of pulses
    tbounds = get_bounds_per_region(tdense, yenvdense)
    nws = {k: v.shape[0] for k, v in tbounds.items()}
    k, npulses = max(nws, key=nws.get), max(nws.values())

    # In cases of multiple pulses, extract PRF and log-spectrum value at PRF
    if npulses > 1:
        logger.debug(f'identifying PRF from time difference between first 2 {k} onsets')
        PRF = np.round(1 / (tbounds[k][1, 0] - tbounds[k][0, 0]), 2)
        PRF_val = np.interp(PRF, freqs, ps_decibel)
    else:
        if get_PRF_val:
            raise ValueError('Cannot compute PRF log-spectrum value for single-pulse waveform')
        PRF = None

    # Mark carrier and PRF frequencies, if specified
    if mark_freqs:
        ax.axvline(Fdrive, c='k', ls='--', label='carrier')
        if PRF is not None:
            ax.axvline(PRF, c='r', ls='--', label='PRF')
    
    # Add legend
    if plot:
        ax.legend()

    # Return figure handle (and log-spectrum value if requested)
    if get_PRF_val:
        return fig, PRF_val
    else:
        return fig


# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-08-08 10:11:50
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-17 10:54:49

from instrulink.camera import grab_camera, TriggerSource

# Grab camera
cam = grab_camera()

# Define acquisition settings
duration = 10.  # s
nacqs = 10
output_fname = 'D:/Theo/testacq/test_video.mp4'

# Acquire videos
try:
    cam.acquire(
        output_fname, duration, nacqs=nacqs, 
        trigger_source=TriggerSource.EXTERNAL)
except KeyboardInterrupt as err:
    cam.stopCapture()

# Disconnect camera
cam.disconnect()
